import logging
import subprocess
import typing as t
from dataclasses import dataclass
from pathlib import Path

logger = logging.getLogger(__name__)


@dataclass
class CMDOutput:
    output: str
    status: t.Literal["success", "fail"] = "success"


def execute(command: list[str], path: str | Path | None = None):
    logging.debug("[CMD]: Running %s.", command)
    result = subprocess.run(command, cwd=path, shell=False, text=True, capture_output=True)

    if result.returncode == 0:
        return CMDOutput(output=result.stdout.strip())
    else:
        logging.warning("[CMD]: Failed running %s:\n%s", command, result.stderr.strip())
        return CMDOutput(output=result.stderr.strip(), status="fail")



from .base import History
from .in_file import InFileHistory
from .in_memory import InMemoryHistory
from .types import AutomatonRunResult

__all__ = [
    "AutomatonRunResult",
    "History",
    "InMemoryHistory",
    "InFileHistory",
]



from automyte import File, Filter, OSFile


class ExampleFilter1(Filter):
    def filter(self, file: File):
        return file.contains("1")


class ExampleFilter2(Filter):
    def filter(self, file: File):
        return file.contains("2")


class ExampleFilter3(Filter):
    def filter(self, file: File):
        return file.contains("3")


class TestFilter:
    def test_logical_and_results_in_file_having_to_pass_both_filters_conditions(self, tmp_os_file):
        file1: OSFile = tmp_os_file("11111")
        file2: OSFile = tmp_os_file("22222")
        file3: OSFile = tmp_os_file("3333")
        correct_file: OSFile = tmp_os_file("111222")

        filter = ExampleFilter1() & ExampleFilter2()

        assert not filter.filter(file=file1)
        assert not filter.filter(file=file2)
        assert not filter.filter(file=file3)
        assert filter.filter(file=correct_file)

    def test_logical_or_results_in_file_having_to_pass_either_filters_conditions(self, tmp_os_file):
        file1: OSFile = tmp_os_file("11111")
        file2: OSFile = tmp_os_file("22222")
        file3: OSFile = tmp_os_file("111222")
        incorrect_file: OSFile = tmp_os_file("3333")

        filter = ExampleFilter1() | ExampleFilter2()

        assert filter.filter(file=file1)
        assert filter.filter(file=file2)
        assert filter.filter(file=file3)
        assert not filter.filter(file=incorrect_file)

    def test_logical_not_results_in_file_having_to_fail_filter_condition(self, tmp_os_file):
        file1: OSFile = tmp_os_file("11111")
        file2: OSFile = tmp_os_file("2")
        filter = ~ExampleFilter1()

        assert not filter.filter(file=file1)
        assert filter.filter(file=file2)

    def test_complex_scenario(self, tmp_os_file):
        file: OSFile = tmp_os_file("111222333")

        filter1 = ExampleFilter1() & ExampleFilter2()
        filter2 = (ExampleFilter1() & ExampleFilter2()) & ~ExampleFilter3()

        assert filter1.filter(file=file)
        assert not filter2.filter(file=file)



from automyte.utils import bash


def test_plain_execution():
    result = bash.execute(["echo", "stdout test"])
    assert result.status == "success"
    assert result.output == "stdout test"


def test_failure_execution():
    result = bash.execute(["ls", "/kladjslkajdslkjasd"])
    assert result.status == "fail"
    assert result.output == "ls: /kladjslkajdslkjasd: No such file or directory"



import typing as t

from automyte.config import Config
from automyte.history import AutomatonRunResult, History, InMemoryHistory
from automyte.project import Project, ProjectURI

from .flow import TasksFlow
from .run_context import RunContext
from .types import FileTask


class Automaton:
    def __init__(
        self,
        name: str,
        projects: list[Project | ProjectURI],
        tasks: TasksFlow | list[FileTask],
        config: Config | None = None,
        history: History | None = None,
    ):
        self.name = name
        self.config: Config = config or Config.get_default()
        self.history: History = history or InMemoryHistory()

        self.projects = []
        for project in projects:
            if isinstance(project, str):
                self.projects.append(Project.from_uri(project))
            else:
                self.projects.append(project)

        if isinstance(tasks, TasksFlow):
            self.flow = tasks
        else:
            self.flow = TasksFlow(*tasks)

    def run(self):
        for project in self._get_target_projects():
            result = AutomatonRunResult(status="running")
            previous_result = self.history.get_status(self.name, project.project_id)
            ctx = RunContext(
                automaton_name=self.name,
                config=self.config,
                vcs=project.vcs,
                project=project,
                current_status=result,
                previous_status=previous_result,
                global_tasks_returns=[],
                file_tasks_returns=[],
            )

            try:
                result = self._execute_for_project(project, ctx)

            except Exception as e:
                result = AutomatonRunResult(status="fail", error=str(e))

            finally:
                self._update_history(project, result)

            if self.config.stop_on_fail and result.status == "fail":
                break

    def _get_target_projects(self) -> t.Generator[Project, None, None]:
        targets = {p.project_id: p for p in self.projects}
        projects_in_history = self.history.read(self.name)
        filter_by_status = lambda status: {  # Get projects from targets based on their status in history.
            proj_id: targets[proj_id] for proj_id, run in projects_in_history.items() if run.status == status
        }

        match self.config.target:
            case "all":
                pass
            case "new":
                new_targets = filter_by_status("new")
                targets = {  # Including projects that are not in the history yet (if added for 2nd run for example).
                    **new_targets,
                    **{pid: proj for pid, proj in targets.items() if pid not in projects_in_history},
                }
            case "failed":
                targets = filter_by_status("fail")
            case "successful":
                targets = filter_by_status("success")
            case "skipped":
                targets = filter_by_status("skipped")
            case _:  # Passed target_id explicitly.
                targets = {pid: proj for pid, proj in targets.items() if pid == self.config.target}

        for project in targets.values():
            yield project

    def _execute_for_project(self, project: Project, ctx: RunContext) -> AutomatonRunResult:
        with project.in_working_state(ctx.config):
            result = self.flow.execute(project=project, ctx=ctx)

        return result

    def _update_history(self, project: Project, result: AutomatonRunResult):
        self.history.set_status(automaton_name=self.name, project_id=project.project_id, status=result)



from pathlib import Path


def parse_dir(directory: str | Path) -> Path:
    if isinstance(directory, str):
        path = Path(directory).expanduser()
    else:
        path = directory.expanduser()

    if not path.is_dir():
        raise ValueError(f"Received invalid directory: {directory}.")

    return path



from unittest.mock import ANY, patch

import pytest

from automyte import Project, TasksFlow
from automyte.automaton.run_context import RunContext
from automyte.automaton.types import TaskReturn
from automyte.discovery.file.base import File
from automyte.discovery.file.os_file import OSFile
from automyte.history.types import AutomatonRunResult


def dummy_task(ctx: RunContext, file: File | None):
    return


def dummy_task2(ctx: RunContext, file: File | None):
    return


class TestFlowInit:
    def test_correctly_sets_tasks(self):
        flow = TasksFlow(
            dummy_task,
            dummy_task,
            [dummy_task, dummy_task],
            preprocess=[dummy_task, dummy_task],
            postprocess=[dummy_task, dummy_task],
        )

        assert len(flow.tasks) == 4
        assert len(flow.preprocess_tasks) == 2
        assert len(flow.postprocess_tasks) == 2

    def test_tasks_are_ordered_correctly(self):
        tasks = iter(TasksFlow(dummy_task, dummy_task2, [dummy_task2, dummy_task], dummy_task2).tasks)

        assert next(tasks) is dummy_task
        assert next(tasks) is dummy_task2
        assert next(tasks) is dummy_task2
        assert next(tasks) is dummy_task
        assert next(tasks) is dummy_task2


class TestFlowExecute:
    def test_tasks_calls_ordering(self, tmp_local_project, run_ctx):
        outputs = []
        preprocess = lambda ctx, file: outputs.append("preprocess")
        postprocess = lambda ctx, file: outputs.append("postprocess")
        normal_task = lambda ctx, file: outputs.append("normal task")
        ctx = run_ctx(dir=tmp_local_project({"hello.txt": "hello flow"}))

        TasksFlow([normal_task], preprocess=[preprocess], postprocess=[postprocess]).execute(ctx.project, ctx)

        assert outputs[0] == "preprocess"
        assert outputs[1] == "normal task"
        assert outputs[2] == "postprocess"

    def test_saves_task_returns_to_ctx(self, tmp_local_project, run_ctx):
        expected_return = TaskReturn(instruction="continue", status="processed", value="smth")
        task = lambda ctx, file: expected_return
        ctx = run_ctx(dir=tmp_local_project({"hello.txt": "hello flow"}))
        with patch("automyte.automaton.run_context.RunContext.save_task_result") as save_mock:
            TasksFlow(task).execute(ctx.project, ctx)

        save_mock.assert_called_once_with(result=expected_return, file=ANY)

    def test_plain_return_is_wrapped_as_task_return_with_correct_defaults(self, tmp_local_project, run_ctx):
        task = lambda ctx, file: "smth"
        ctx = run_ctx(dir=tmp_local_project({"hello.txt": "hello flow"}))

        with patch("automyte.automaton.run_context.RunContext.save_task_result") as save_mock:
            TasksFlow(task).execute(ctx.project, ctx)

        save_mock.assert_called_once_with(
            result=TaskReturn(value="smth", instruction="continue", status="processed"), file=ANY
        )

    def test_tasks_are_called_once_per_file(self, tmp_local_project, run_ctx):
        files_called_for = []
        track_calls = lambda ctx, file: files_called_for.append(file.name)
        ctx = run_ctx(dir=tmp_local_project({"src": {"hello.txt": "hello flow", "bye.txt": "bye flow"}}))

        TasksFlow(track_calls).execute(ctx.project, ctx)

        assert len(files_called_for) == 2
        assert next(f for f in files_called_for if f == "hello.txt")
        assert next(f for f in files_called_for if f == "bye.txt")

    def test_changes_are_applied_before_postprocess_tasks(self, tmp_local_project, run_ctx):
        ctx = run_ctx(dir=tmp_local_project({"src": {"hello.txt": "hello flow"}}))
        update_file = lambda ctx, file: file.edit("changed")
        read_file = lambda ctx, file: OSFile(fullname=f"{ctx.project.rootdir}/src/hello.txt").get_contents()

        TasksFlow(update_file, postprocess=[read_file]).execute(ctx.project, ctx)

        assert ctx.previous_return.value == "changed"

    def test_abort_task_instructions_return_immediately(self, tmp_local_project, run_ctx):
        ctx = run_ctx(dir=tmp_local_project({"src": {"hello.txt": "hello flow"}}))
        has_next_task_been_called = []
        check_if_called = lambda ctx, file: has_next_task_been_called.append("whatever, will just check falsy list")
        instruct = lambda ctx, file: TaskReturn(instruction="abort", value="oops")

        output = TasksFlow(instruct, check_if_called).execute(ctx.project, ctx)

        assert not has_next_task_been_called
        assert output == AutomatonRunResult(status="fail", error="oops")

    def test_skip_task_instructions_return_immediately(self, tmp_local_project, run_ctx):
        ctx = run_ctx(dir=tmp_local_project({"src": {"hello.txt": "hello flow"}}))
        has_next_task_been_called = []
        check_if_called = lambda ctx, file: has_next_task_been_called.append("whatever, will just check falsy list")
        instruct = lambda ctx, file: TaskReturn(instruction="skip")

        output = TasksFlow(instruct, check_if_called).execute(ctx.project, ctx)

        assert not has_next_task_been_called
        assert output == AutomatonRunResult(status="skipped")

    def test_exception_raised_inside_task_results_in_abort(self, tmp_local_project, run_ctx):
        ctx = run_ctx(dir=tmp_local_project({"src": {"hello.txt": "hello flow"}}))
        has_next_task_been_called = []
        check_if_called = lambda ctx, file: has_next_task_been_called.append("whatever, will just check falsy list")

        def failure_of_a_task(ctx, file):
            raise Exception("oops")

        output = TasksFlow(failure_of_a_task, check_if_called).execute(ctx.project, ctx)

        assert not has_next_task_been_called
        assert output == AutomatonRunResult(status="fail", error="oops")



from __future__ import annotations

from ..file import File


class Filter:
    def filter(self, file: File) -> bool:
        raise NotImplementedError

    def __call__(self, file: File) -> bool:
        return self.filter(file=file)

    def __and__(self, other: "Filter") -> "Filter":
        """Returns a filter that represents logical AND."""

        class AndFilter(Filter):
            def __init__(self, filter1: Filter, filter2: Filter):
                self.filter1 = filter1
                self.filter2 = filter2

            def filter(self, file: File) -> bool:
                return self.filter1.filter(file) and self.filter2.filter(file)

        return AndFilter(self, other)

    def __or__(self, other: "Filter") -> "Filter":
        """Returns a filter that represents logical OR."""

        class OrFilter(Filter):
            def __init__(self, filter1: Filter, filter2: Filter):
                self.filter1 = filter1
                self.filter2 = filter2

            def filter(self, file: File) -> bool:
                return self.filter1.filter(file) or self.filter2.filter(file)

        return OrFilter(self, other)

    def __invert__(self) -> "Filter":
        class NotFilter(Filter):
            def __init__(self, filter: Filter) -> None:
                self.inverted_filter = filter

            def filter(self, file: File) -> bool:
                return not self.inverted_filter.filter(file)

        return NotFilter(self)



import abc
import typing as t

from ..file import File


# NOTE: Maybe split it into FilesBackend + ProjectExplorer class, so then ProjectExplorer is responsible for filters, backend is for getting/saving files
class ProjectExplorer(abc.ABC):
    def get_rootdir(self) -> str:
        """To be overriden by child classes to provide project/vcs with access to explorer's rootdir."""
        raise NotImplementedError

    def set_rootdir(self, newdir: str) -> str:
        """To be overriden by child classes to allow project to adjust explorer dir during setup() phase."""
        raise NotImplementedError

    def explore(self) -> t.Generator[File, None, None]:
        """To be inherited from and override accessing/saving project's files logic"""
        raise NotImplementedError

    def flush(self):
        """Centralised hook to actually apply all necessary changes for all files that require it."""
        raise NotImplementedError



import re

from automyte import (
    Automaton,
    AutomatonRunResult,
    Config,
    ContainsFilter,
    File,
    InMemoryHistory,
    LocalFilesExplorer,
    Project,
    RunContext,
    TaskReturn,
    TasksFlow,
    guards,
)


def instruct_to_skip(ctx: RunContext, file: File):
    return TaskReturn(instruction="skip")


def instruct_to_abort(ctx: RunContext, file: File):
    return TaskReturn(instruction="abort", value="smth is wrong")


def error_out(ctx: RunContext, file: File):
    raise Exception("oops")


def second_task(ctx: RunContext, file: File):
    file.edit(re.sub(r"world", "there", file.get_contents()))


def test_skip_works_immediately_and_updates_history(tmp_local_project):
    dir = tmp_local_project(structure={"src": {"hello.txt": "hello world!"}})

    history = InMemoryHistory()
    Automaton(
        name="impl1",
        config=Config.get_default().set_vcs(dont_disrupt_prior_state=False),
        history=history,
        projects=[
            Project(
                project_id="test_project",
                explorer=LocalFilesExplorer(rootdir=dir, filter_by=ContainsFilter(contains="hello world")),
            )
        ],
        tasks=TasksFlow([instruct_to_skip, second_task]),
    ).run()

    with open(f"{dir}/src/hello.txt", "r") as f:
        assert f.read() == "hello world!"  # Make sure it's unchanged.

    assert history.get_status("impl1", "test_project") == AutomatonRunResult(status="skipped")


def test_abort_works_immediately_and_updates_history(tmp_local_project):
    dir = tmp_local_project(structure={"src": {"hello.txt": "hello world!"}})

    history = InMemoryHistory()
    Automaton(
        name="impl1",
        config=Config.get_default().set_vcs(dont_disrupt_prior_state=False),
        history=history,
        projects=[
            Project(
                project_id="test_project",
                explorer=LocalFilesExplorer(rootdir=dir, filter_by=ContainsFilter(contains="hello world")),
            )
        ],
        tasks=TasksFlow([instruct_to_abort, second_task]),
    ).run()

    with open(f"{dir}/src/hello.txt", "r") as f:
        assert f.read() == "hello world!"  # Make sure it's unchanged.

    assert history.get_status("impl1", "test_project") == AutomatonRunResult(status="fail", error="smth is wrong")


def test_task_exception_abort_run(tmp_local_project):
    dir = tmp_local_project(structure={"src": {"hello.txt": "hello world!"}})

    history = InMemoryHistory()
    Automaton(
        name="impl1",
        config=Config.get_default().set_vcs(dont_disrupt_prior_state=False),
        history=history,
        projects=[
            Project(
                project_id="test_project",
                explorer=LocalFilesExplorer(rootdir=dir, filter_by=ContainsFilter(contains="hello world")),
            )
        ],
        tasks=TasksFlow([error_out, second_task]),
    ).run()

    with open(f"{dir}/src/hello.txt", "r") as f:
        assert f.read() == "hello world!"  # Make sure it's unchanged.

    assert history.get_status("impl1", "test_project") == AutomatonRunResult(status="fail", error="oops")



import re

from .base import File, Filter


class ContainsFilter(Filter):
    def __init__(self, contains: str | list[str], regexp: bool = False) -> None:
        self.text = contains if isinstance(contains, list) else [contains]
        self.use_regexp = regexp

    def filter(self, file: File) -> bool:
        if not self.use_regexp:
            return any(file.contains(occurance) for occurance in self.text)

        file_contents = file.get_contents()
        return any(re.search(pattern, file_contents) for pattern in self.text)



import os
import typing as t
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from automyte.automaton.run_context import RunContext
from automyte.config.base import Config
from automyte.discovery import OSFile
from automyte.history.types import AutomatonRunResult
from automyte.project import Project
from automyte.utils import bash
from automyte.utils.random import random_hash
from automyte.vcs.base import VCS

_DirName: t.TypeAlias = str
_FileName: t.TypeAlias = str
_FileContents: t.TypeAlias = str
_ProjectStructure: t.TypeAlias = dict[_DirName | _FileName, t.Union[_FileContents, "_ProjectStructure"]]


@pytest.fixture
def tmp_local_project():
    """Setup a tmp project structure with folders, files and their contents for testing.

    Returns a factory function which accepts dictionary structure where we specify project structure.
    All projects are created in tmp folders which are cleaned up automatically after each run.

    Don't pass "dir" argument, as it is only used for recursive call for internal implementation.

    Example:
        tmp_local_project(structure={
            'src': {
                'subdir1': {
                    'hello.txt': 'this will be the text for src/subdir1/hello.txt file',
                    'bye.py': 'print("good bye")',
                },
                'subdir2': {...},
            ...
            }
        })
    """
    rootdirs = []
    try:

        def _create_tmp_project(structure: _ProjectStructure, dir: str | None = None):
            if dir:  # Recursive call, just need to create child dirs.
                if not Path(dir).exists():
                    os.mkdir(dir)
                current_dir = dir
            else:  # First call, need to create TMP dir as parent and add it to array for removal.
                new_tmp_dir = TemporaryDirectory()
                rootdirs.append(new_tmp_dir)
                current_dir = new_tmp_dir.name

            for name, content in structure.items():
                if isinstance(content, str):  # Encountered a file.
                    with open(Path(current_dir) / name, "w") as f:
                        f.write(content)
                else:  # Encoruntered a folder, so need to generate the whole structure again, recursively.
                    _create_tmp_project(structure=content, dir=f"{current_dir}/{name}")

            return current_dir

        yield _create_tmp_project

    finally:
        for dir in rootdirs:
            dir.cleanup()


@pytest.fixture
def tmp_git_repo(tmp_local_project):
    def _tmp_git_repo_factory(initial_structure, unstaged_structure=None):
        dir = tmp_local_project(initial_structure)
        bash.execute(["git", "-C", dir, "init"])
        bash.execute(["git", "-C", dir, "add", "--all"])
        bash.execute(["git", "-C", dir, "commit", "-m", "Initial commit"])

        if unstaged_structure:
            tmp_local_project(unstaged_structure, dir=dir)

        return dir

    return _tmp_git_repo_factory


@pytest.fixture
def run_ctx():
    def _ctx_factory(
        dir: str | Path, vcs: VCS | None = None, project: Project | None = None, automaton_name: str = "auto"
    ):
        if project is None:
            project = Project(project_id=random_hash(), rootdir=str(dir))
        if vcs is None:
            vcs = project.vcs

        return RunContext(
            automaton_name=automaton_name,
            config=Config.get_default().set_vcs(dont_disrupt_prior_state=False),
            vcs=vcs,
            project=project,
            current_status=AutomatonRunResult("running"),
            previous_status=AutomatonRunResult("new"),
            global_tasks_returns=[],
            file_tasks_returns=[],
        )

    return _ctx_factory


@pytest.fixture
def tmp_os_file(tmp_local_project):
    def _tmp_file_factory(contents: str, filename: str | None = None) -> OSFile:
        filename = filename or random_hash()
        dir = tmp_local_project(structure={filename: contents})

        filepath = Path(dir) / filename
        with open(filepath, "w") as f:
            f.write(contents)

        return OSFile(fullname=str(filepath))

    return _tmp_file_factory



from pathlib import Path

from automyte import (
    Automaton,
    AutomatonRunResult,
    Config,
    ContainsFilter,
    File,
    InMemoryHistory,
    LocalFilesExplorer,
    Project,
    RunContext,
    TaskReturn,
    TasksFlow,
    guards,
)
from automyte.tasks import vcs
from automyte.utils import bash


def lol(ctx: RunContext, file: File):
    import re

    file.edit(re.sub(r"world", "there", file.get_contents()))


def test_worktree_setup(tmp_local_project):
    dir = tmp_local_project(structure={"src": {"hello.txt": "hello world!"}})

    # Worktrees require at least 1 commit
    bash.execute(["git", "-C", dir, "init"])
    bash.execute(["git", "-C", dir, "add", "."])
    bash.execute(["git", "-C", dir, "commit", "-m", "hello"])

    Automaton(
        name="impl1",
        config=Config.get_default(),
        projects=[
            Project(
                project_id="test_project",
                explorer=LocalFilesExplorer(rootdir=dir, filter_by=ContainsFilter(contains="hello world")),
            ),
        ],
        tasks=TasksFlow(
            [
                lol,
                # vcs.add()
            ],
            preprocess=[
                # vcs.run('init')
            ],
            postprocess=[
                vcs.add("."),
                vcs.commit("testing worktree"),
                # Breakpoint(),
            ],
        ),
    ).run()

    assert "src/hello.txt" in bash.execute(["git", "-C", dir, "diff", "master..automate"]).output



from automyte import (
    Automaton,
    AutomatonRunResult,
    Config,
    ContainsFilter,
    File,
    InMemoryHistory,
    LocalFilesExplorer,
    Project,
    RunContext,
    TaskReturn,
    TasksFlow,
    guards,
)


def lol(ctx: RunContext, file: File):
    import re

    file.edit(re.sub(r"world", "there", file.get_contents()))


def test_replacing_text(tmp_local_project):
    dir = tmp_local_project(
        structure={
            "src": {
                "hello.txt": "hello world!",
            },
        }
    )

    Automaton(
        name="impl1",
        config=Config.get_default().set_vcs(dont_disrupt_prior_state=False),
        projects=[
            Project(
                project_id="test_project",
                explorer=LocalFilesExplorer(rootdir=dir, filter_by=ContainsFilter(contains="hello world")),
            ),
        ],
        tasks=TasksFlow([lol]),
    ).run()

    with open(f"{dir}/src/hello.txt", "r") as f:
        assert f.read() == "hello there!"



import contextlib
from dataclasses import dataclass

from automyte.config import Config
from automyte.discovery import File
from automyte.history import AutomatonRunResult
from automyte.project import Project
from automyte.vcs import VCS

# from automyte.tasks import TaskReturn
from .types import TaskReturn


@dataclass
class RunContext:
    automaton_name: str
    config: Config
    vcs: VCS
    project: Project
    current_status: AutomatonRunResult
    previous_status: AutomatonRunResult
    global_tasks_returns: list[TaskReturn]
    file_tasks_returns: list[TaskReturn]
    # NOTE: These fields are not used for now, so not gonna bother implementing them for now.
    # previous_task: BaseTask | None = None
    # next_task: BaseTask | None = None
    # current_file: File | None = None  # None for pre/post process tasks.

    @property
    def previous_return(self):
        """Return previously saved task execution result.

        pre/post process tasks returns are saved indefinitely inside automaton run
        main tasks section returns get cleaned up between each file.
        """
        with contextlib.suppress(IndexError):
            if self.file_tasks_returns:
                return self.file_tasks_returns[-1]
            elif self.global_tasks_returns:
                return self.global_tasks_returns[-1]

        return TaskReturn(instruction="continue", value=None)

    def save_task_result(self, result: TaskReturn, file: File | None):
        if file is None:
            self.global_tasks_returns.append(result)
        else:
            self.file_tasks_returns.append(result)
        return result

    def cleanup_file_returns(self):
        self.file_tasks_returns.clear()



from pathlib import Path
from tempfile import TemporaryDirectory

from automyte import LocalFilesExplorer
from automyte.discovery.file.os_file import File, OSFile
from automyte.discovery.filters.base import Filter


class TestLocalFilesExplorerExplore:
    def test_returns_all_files_in_the_rootdir_if_no_filters_provided(self, tmp_local_project):
        dir = tmp_local_project(
            {"src": {"hello.txt": "hello explorer"}, "upper": {"inner": {"nested.py": "print(123)"}}}
        )

        all_files = list(LocalFilesExplorer(rootdir=dir).explore())

        assert next((f for f in all_files if f.fullpath == Path(dir) / "src" / "hello.txt"), None)
        assert next((f for f in all_files if f.fullpath == Path(dir) / "upper" / "inner" / "nested.py"), None)

    def test_returns_only_files(self, tmp_local_project):
        dir = tmp_local_project(
            {"src": {"hello.txt": "hello explorer"}, "upper": {"inner": {"nested.py": "print(123)"}}}
        )

        all_files = list(LocalFilesExplorer(rootdir=dir).explore())
        assert len(all_files) == 2

    def test_filters_files_accordingly_to_provided_filter(self, tmp_local_project):
        dir = tmp_local_project(
            {"src": {"hello.txt": "hello explorer"}, "upper": {"inner": {"nested.py": "print(123)"}}}
        )

        class TMPFilter(Filter):
            def filter(self, file: File) -> bool:
                self.has_been_called = True
                return file.name == "hello.txt"

        filter = TMPFilter()
        all_files = list(LocalFilesExplorer(rootdir=dir, filter_by=filter).explore())

        assert all_files[0].name == "hello.txt"
        assert len(all_files) == 1
        assert filter.has_been_called


class TestLocalFilesExplorerRootdir:
    def test_get_rootdir_returns_proper_field(self):
        with TemporaryDirectory() as dir:
            assert LocalFilesExplorer(rootdir=dir).get_rootdir() == dir

    def test_set_rootdir_updates_it(self):
        with TemporaryDirectory() as dir:
            explorer = LocalFilesExplorer(rootdir=dir)
            explorer.set_rootdir(newdir=f"{dir}/new")
            assert explorer.get_rootdir() == f"{dir}/new"


class TestLocalFilesExplorerFlush:
    def test_flushes_tainted_files(self, tmp_local_project):
        dir = tmp_local_project(
            {"src": {"hello.txt": "hello explorer"}, "upper": {"inner": {"nested.py": "print(123)"}}}
        )
        explorer = LocalFilesExplorer(rootdir=dir)
        for file in explorer.explore():
            if file.name == "hello.txt":
                file = file.edit("good bye explorer")

        explorer.flush()

        with open(f"{dir}/src/hello.txt", "r") as flushed_file:
            assert flushed_file.read() == "good bye explorer"

    def test_doesnt_change_untainted_files(self, tmp_local_project):
        dir = tmp_local_project(
            {"src": {"hello.txt": "hello explorer"}, "upper": {"inner": {"nested.py": "print(123)"}}}
        )
        explorer = LocalFilesExplorer(rootdir=dir)
        for file in explorer.explore():
            if file.name == "hello.txt":
                file = file.edit("good bye explorer")

        explorer.flush()

        with open(f"{dir}/upper/inner/nested.py", "r") as nested_file:
            assert nested_file.read() == "print(123)"


class TestLocalFilesExplorerIgnoreUtilFiles:
    def test_ignores_files_by_default(self, tmp_git_repo):
        dir = tmp_git_repo({"src": {"hello.txt": "hello there", "node_modules": {"bun.txt": "rabbit"}}})

        files = list(LocalFilesExplorer(rootdir=dir).explore())

        assert len(files) == 1
        assert files[0].name == "hello.txt"

    def test_ignore_can_be_turned_off(self, tmp_git_repo):
        dir = tmp_git_repo({"src": {"hello.txt": "hello there", "node_modules": {"bun.txt": "rabbit"}}})

        files = list(LocalFilesExplorer(rootdir=dir, ignore_locations=[]).explore())

        assert len(files) > 1
        assert next(f for f in files if f.name == "bun.txt")
        assert next(f for f in files if ".git" in f.folder)



import contextlib
from pathlib import Path
from unittest.mock import patch

import pytest

from automyte import Project
from automyte.config.base import Config
from automyte.discovery.explorers.base import ProjectExplorer
from automyte.discovery.explorers.local_files import LocalFilesExplorer
from automyte.vcs.base import VCS
from automyte.vcs.git import Git


class DummyExplorer(ProjectExplorer):
    def get_rootdir(self) -> str:
        return "smth"

    def set_rootdir(self, newdir: str):
        return "smth"

    def flush(self):
        self.flushed = True


class DummyVCS(VCS):
    def run(self, *args):
        return

    @contextlib.contextmanager
    def preserve_state(self, config):
        self.preserve_state_called = True
        yield "newdir"


class TestProjectInit:
    def test_requires_at_least_one_of_rootdir_or_explorer(self):
        with pytest.raises(ValueError):
            Project(project_id="proj1", vcs=Git(""))

    def test_uses_local_files_explorer_by_default(self, tmp_local_project):
        dir = tmp_local_project({})
        explorer = Project("proj1", rootdir=dir).explorer

        assert isinstance(explorer, LocalFilesExplorer)
        assert explorer.get_rootdir() == dir

    def test_assigns_roodir_from_explorer_if_no_rootdir_is_passed_in_init(self):
        explorer = DummyExplorer()

        proj = Project(project_id="proj1", explorer=explorer)

        assert proj.rootdir == "smth"

    def test_uses_git_vcs_with_correct_rootdir_by_default(self, tmp_local_project):
        dir = tmp_local_project({})
        vcs = Project(project_id="proj1", rootdir=dir).vcs

        assert isinstance(vcs, Git)
        assert vcs.original_rootdir == dir

    def test_validates_rootdir(self):
        with pytest.raises(ValueError):
            Project("proj1", rootdir="/some/definitely/not/existant/dir")


class TestProjectApplyChanges:
    def test_calls_explorer_flush(self):
        explorer = DummyExplorer()

        Project(project_id="proj1", explorer=explorer).apply_changes()

        assert explorer.flushed


class TestProjectInWorkingState:
    def test_calls_vcs_preserve_state(self, tmp_local_project):
        dir = tmp_local_project({})
        vcs = DummyVCS()

        with Project("proj1", rootdir=dir, vcs=vcs).in_working_state(Config.get_default()):
            ...

        assert vcs.preserve_state_called

    def test_sets_and_resets_own_rootdir_if_necessary(self, tmp_local_project):
        dir = tmp_local_project({})
        vcs = DummyVCS()
        project = Project("proj1", rootdir=dir, vcs=vcs)

        with project.in_working_state(Config.get_default()):
            assert project.rootdir == "newdir"
        assert project.rootdir == dir

    def test_updates_explorer_rootdir(self, tmp_local_project):
        dir = tmp_local_project({})
        vcs = DummyVCS()
        project = Project("proj1", rootdir=dir, vcs=vcs)

        with project.in_working_state(Config.get_default()):
            assert project.explorer.get_rootdir() == "newdir"
        assert project.explorer.get_rootdir() == dir


class TestProjectFromUri:
    def test_generates_correct_rootdir(self, tmp_local_project):
        dir = tmp_local_project({})
        assert Project.from_uri(dir).rootdir == dir

    def test_generates_unique_but_readable_project_id(self, tmp_local_project):
        dir = tmp_local_project({})
        with patch("automyte.project.project.random_hash", return_value="smth"):
            project = Project.from_uri(dir)

            assert project.project_id == f"smth_{Path(dir).name}"

    def test_sets_correct_explorer(self, tmp_local_project):
        dir = tmp_local_project({})
        explorer = Project.from_uri(dir).explorer

        assert isinstance(explorer, LocalFilesExplorer)
        assert explorer.get_rootdir() == dir



from automyte.automaton import RunContext
from automyte.discovery import OSFile
from automyte.tasks.utils import fs


class TestFileSystemFlush:
    def test_calls_file_flush_when_provided(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("whatever").edit("test flushing")
        ctx: RunContext = run_ctx(file.folder)

        fs.flush()(ctx, file)

        with open(file.fullpath, "r") as disk_file:
            assert disk_file.read() == "test flushing"



from __future__ import annotations

import abc
import typing as t


class File(abc.ABC):
    @property
    def folder(self) -> str:
        raise NotImplementedError

    @property
    def name(self) -> str:
        raise NotImplementedError

    def get(self) -> t.Self:
        raise NotImplementedError

    def flush(self) -> None:
        raise NotImplementedError

    def contains(self, text: str) -> bool:
        raise NotImplementedError

    def move(self, to: str | None = None, new_name: str | None = None) -> File:
        raise NotImplementedError

    def get_contents(self) -> str:
        raise NotImplementedError

    def edit(self, text: str) -> File:
        raise NotImplementedError

    def delete(self) -> File:
        raise NotImplementedError

    @property
    def is_tainted(self) -> bool:
        raise NotImplementedError



import typing as t

from automyte.discovery import File
from automyte.history import AutomatonRunResult
from automyte.project import Project

from .run_context import RunContext
from .types import BaseTask, FileTask, InstructionForAutomaton, TaskReturn


class TasksFlow:
    def __init__(
        self,
        *tasks: FileTask | list[FileTask],
        preprocess: list[BaseTask] | None = None,
        postprocess: list[BaseTask] | None = None,
    ):
        self.preprocess_tasks = preprocess or []
        self.postprocess_tasks = postprocess or []

        self.tasks = []
        for task in tasks:
            if isinstance(task, list):
                self.tasks.extend(task)
            else:
                self.tasks.append(task)

    def execute(self, project: Project, ctx: "RunContext"):
        for preprocess_task in self.preprocess_tasks:
            result = execute_task(ctx=ctx, task=preprocess_task, file=None)
            if result:
                return result

        for file in project.explorer.explore():
            for process_file_task in self.tasks:
                result = execute_task(ctx=ctx, task=process_file_task, file=file)
                if result:
                    return result

            ctx.cleanup_file_returns()

        # Has to be called prior to postprocess tasks, otherwise files changes are not reflected on disk before vcs calls.
        project.apply_changes()

        for post_task in self.postprocess_tasks:
            result = execute_task(ctx=ctx, task=post_task, file=None)
            if result:
                return result

        return AutomatonRunResult(status="success")


def execute_tasks_sequence(tasks: list[BaseTask], ctx: RunContext, file: File | None):
    """Similar to 'execute_task', except designed to run a sequence of tasks.

    This function is necessary to stop calls chain, if any of the tasks results in 'abort' or 'skip' instruction.
    """
    instruction: InstructionForAutomaton = "continue"
    for task in tasks:
        instruction = handle_task_call(task=task, ctx=ctx, file=file)
        # If any of the tasks in the sequence resulted in 'abort' or 'skip',
        # don't execute any of the remaining tasks and stop automaton run.
        if instruction != "continue":
            break

    if instruction == "skip":
        return AutomatonRunResult(status="skipped")
    elif instruction == "abort":
        return AutomatonRunResult(status="fail", error=str(ctx.previous_return.value))

    return None


def execute_task(ctx: "RunContext", task: BaseTask, file: File | None) -> AutomatonRunResult | None:
    """The main entrypoint for actually calling tasks inside automaton.

    Will wrap all returns into TaskReturns and save them to ctx;
    Handles exceptions inside tasks and sends abort instruction to automaton;

    If return value is None - flow is to just continue executing tasks,
        otherwise, need to interrupt.
    """
    instruction = handle_task_call(ctx=ctx, task=task, file=file)

    if instruction == "skip":
        return AutomatonRunResult(status="skipped")
    elif instruction == "abort":
        return AutomatonRunResult(status="fail", error=str(ctx.previous_return.value))

    return None


def handle_task_call(ctx: "RunContext", task: BaseTask, file: File | None) -> "InstructionForAutomaton":
    """Convenience wrapper for calling and handling all tasks.

    Wraps plain python values into TaskReturns,
        so that user doesn't have to do it unless they want to specify behaviour;
    Save task return into ctx;
    Return instruction for automaton on what to do next (like skip the project, continue or abort right away).

    If the task raised an Exception - save it's value into task return with "errored" status and instruct to abort.
    """
    try:
        task_result = wrap_task_result(task(ctx, file))

    except Exception as e:
        ctx.save_task_result(result=TaskReturn(instruction="abort", value=str(e), status="errored"), file=file)
        return "abort"

    else:
        ctx.save_task_result(result=task_result, file=file)
        return task_result.instruction


def wrap_task_result(value: t.Any) -> TaskReturn:
    if isinstance(value, TaskReturn):
        return value
    else:
        return TaskReturn(instruction="continue", status="processed", value=value)



import pdb
import typing as t

from automyte.automaton import RunContext
from automyte.discovery import File

_user_prompt = """
Please select one of the following options:
(d) - Debug (spawn pdb.set_trace() with access to task context, like ctx and file)
(a) - Apply changes (immediately flushes all files to disk, useful to inspect currect project state)
(p) - Print current project working directory
(c) - Continue (exit breakpoint, continue automaton execution)

Any options besides (c) will keep breakpoint active.
Enter your choice (default: "c"):
"""
_allowed_options = ["d", "a", "p", "c"]
_possible_options = t.Literal["d", "a", "p", "c"]


# TODO: Update implementation, to have "interactive=True, call_once=True" in init
# call_once is for it to only be called once, either on first or if possible - last file
# interactive - enables consoles input, otherwise - allow calling Breakpoint.flush(), ...?
class Breakpoint:
    def __init__(self, select_option: _possible_options | None = None):
        # self.select_option = select_option
        ...

    def __call__(self, ctx: RunContext, file: File | None):
        while True:
            choice = self._get_input(_user_prompt)

            if not choice:
                return
            elif choice not in _allowed_options:
                print(f"Invalid option selected, please enter one of: {_allowed_options}")
            else:
                match t.cast(_possible_options, choice):
                    case "c":
                        return
                    case "d":
                        pdb.set_trace()
                        return
                    case "a":
                        ctx.project.apply_changes()
                        print("Changed applied, you can inspect project state and then continue...")
                    case "p":
                        print(ctx.project.rootdir)

    def _get_input(self, text: str):
        return input(text)



import abc

from .types import AutomatonRunResult, ProjectID


class History(abc.ABC):
    def set_status(self, automaton_name: str, project_id: str, status: AutomatonRunResult):
        raise NotImplementedError

    def get_status(self, automaton_name: str, project_id: str) -> AutomatonRunResult:
        raise NotImplementedError

    def read(self, automaton_name: str) -> dict[ProjectID, AutomatonRunResult]:
        """Return all project's history status

        If <automaton_name> automaton has never ran for a project - return status as "new".
        """
        raise NotImplementedError



from __future__ import annotations

import abc
import contextlib
import typing as t
from pathlib import Path

from automyte.config import VCSConfig
from automyte.config.vcs import SupportedVCS
from automyte.discovery import File, Filter


class VCSException(Exception):
    """To be raised in case if any actual vcs cli command fails, or you want to interrupt automaton execution."""


class VCS(abc.ABC):
    """Class used for 2 purposes: providing means of local project state preservation
        and 'run' util for vcs cli interface

    'preserve_state' is a context manager which allows automatons not to interfere with a local working state
    of a project (if a dev was working on some feature - we don't want to simply discard the changes)

    'run' method basically accepts args as a bash executable array of commands to call any vcs command
    of arbitrary difficulty; it will call those commands inside project's current working dir.

    NOTE: VCS tasks in contrib or core utils are to rely on RunContext to get access to project rootdir and stuff
        and then call corresponding vcs commands or directly call 'run' method for any vcs command with flags, etc.
    """

    @contextlib.contextmanager
    def preserve_state(self, config: VCSConfig):
        raise NotImplementedError

    def run(self, *subcommand_with_flags):
        raise NotImplementedError


class VCSCmdBuilder:
    def __init__(self, vcs: SupportedVCS):
        self._vcs = vcs
        self._command = ""
        self._command_args = []
        self._dir = None

    def cmd(self, vcs_command: str):
        self._command = vcs_command
        return self

    def args(self, *args: str) -> "VCSCmdBuilder":
        self._command_args.extend(list(args))
        return self

    def in_dir(self, dir: Path | str) -> "VCSCmdBuilder":
        self._dir = str(dir)
        return self

    def to_cmd(self):
        result = [self._vcs]

        if self._dir:
            if self._vcs == "git":
                result.extend(["-C", self._dir])
            else:
                raise ValueError("Add support for new VCS systems.")

        result.append(self._command)
        result.extend(self._command_args)

        return result



from automyte import ContainsFilter


class TestContainsFilter:
    def test_single_text_returns_correctly(self, tmp_os_file):
        file = tmp_os_file("substring")

        assert ContainsFilter(contains="sub").filter(file=file)
        assert not ContainsFilter(contains="text").filter(file=file)

    def test_array_value_in_contains_filters_by_at_least_one_match(self, tmp_os_file):
        file = tmp_os_file("substring")

        assert ContainsFilter(contains=["sub", "text"]).filter(file=file)
        assert not ContainsFilter(contains=["random", "text"]).filter(file=file)

    def test_regexp_returns_correctly_for_str_contains(self, tmp_os_file):
        file = tmp_os_file("substring")

        assert ContainsFilter(contains=r"su.string", regexp=True).filter(file=file)
        assert not ContainsFilter(contains=r"si.*", regexp=True).filter(file=file)

    def test_array_value_in_contains_with_regexp_filters_by_at_least_one_match(self, tmp_os_file):
        file = tmp_os_file("substring")

        assert ContainsFilter(contains=[r"su.string", r"text"], regexp=True).filter(file=file)
        assert not ContainsFilter(contains=[r"si.*", r"what.*"], regexp=True).filter(file=file)



from pathlib import Path
from unittest.mock import patch

import pytest

from automyte import Config, Git
from automyte.utils import bash


class TestGitRun:
    def test_correctly_generates_prompt(self, tmp_local_project):
        dir = tmp_local_project({"src": {"hello.txt": "hello there"}})

        with patch("automyte.utils.bash.execute") as mock_execute:
            Git(rootdir=dir).run("commit", "--amend", "--no-edit")
            mock_execute.assert_called_once_with(["git", "-C", dir, "commit", "--amend", "--no-edit"])

    def test_actually_runs_command(self, tmp_git_repo):
        dir = tmp_git_repo(
            initial_structure={"src": {"hello.txt": "hello git"}}, unstaged_structure={"src": {"hello.txt": ""}}
        )

        Git(rootdir=dir).run("add", "src")

        staged_files_diff = bash.execute(["git", "-C", dir, "diff", "--name-only", "--cached"]).output
        assert "src/hello.txt" in staged_files_diff


class TestGitPreserveState:
    def test_doesnt_disrupt_user_state(self, tmp_git_repo):
        dir = tmp_git_repo(
            initial_structure={"src": {"hello.txt": "hello git"}},
            unstaged_structure={"src": {"hello.txt": "don't touch me"}},
        )

        with Git(rootdir=dir).preserve_state(Config.get_default().set_vcs(work_branch="verify").vcs) as new_workdir:
            # Running commands inside new workdir, not touching old state.

            bash.execute(["git", "-C", new_workdir, "add", "src"])
            bash.execute(["git", "-C", new_workdir, "commit", "-m", "commit in worktree"])

            staged_files = bash.execute(["git", "-C", dir, "diff", "--name-only", "--cached"]).output
            assert not staged_files
            commits_count = len(bash.execute(["git", "-C", dir, "log", "--oneline"]).output.splitlines())
            assert commits_count == 1

        # Also verifying after preserve_state has exited.
        staged_files = bash.execute(["git", "-C", dir, "diff", "--name-only", "--cached"]).output
        assert not staged_files
        commits_count = len(bash.execute(["git", "-C", dir, "log", "--oneline"]).output.splitlines())
        assert commits_count == 1

    def test_creates_new_worktree(self, tmp_git_repo):
        dir = tmp_git_repo(initial_structure={"src": {"hello.txt": "hello git"}})
        initial_worktrees_count = len(bash.execute(["git", "-C", dir, "worktree", "list"]).output.splitlines())

        with Git(rootdir=dir).preserve_state(Config.get_default().set_vcs(work_branch="verify").vcs):
            worktrees_list = bash.execute(["git", "-C", dir, "worktree", "list"]).output
            assert "[verify]" in worktrees_list
            assert len(worktrees_list.split("\n")) == initial_worktrees_count + 1

    def test_removes_worktree_after_exit(self, tmp_git_repo):
        dir = tmp_git_repo(initial_structure={"src": {"hello.txt": "hello git"}})
        initial_worktrees_count = len(bash.execute(["git", "-C", dir, "worktree", "list"]).output.split("\n"))

        with Git(rootdir=dir).preserve_state(Config.get_default().set_vcs(work_branch="verify").vcs):
            ...

        worktrees_list = bash.execute(["git", "-C", dir, "worktree", "list"]).output.splitlines()
        assert len(worktrees_list) == initial_worktrees_count

    def test_creates_worktree_in_a_child_dir(self, tmp_git_repo):
        dir = tmp_git_repo(initial_structure={"src": {"hello.txt": "hello git"}})

        with Git(rootdir=dir).preserve_state(Config.get_default().set_vcs(work_branch="verify").vcs) as new_workdir:
            worktree_desc = bash.execute(["git", "-C", dir, "worktree", "list"]).output.splitlines()[1]
            assert Path(new_workdir).resolve().is_relative_to(Path(dir).resolve())
            assert new_workdir in worktree_desc

    def test_updates_and_then_resets_own_workdir(self, tmp_git_repo):
        dir = tmp_git_repo(initial_structure={"src": {"hello.txt": "hello git"}})
        git = Git(rootdir=dir)
        original_workdir = git.workdir

        with git.preserve_state(Config.get_default().set_vcs(work_branch="verify").vcs) as new_workdir:
            assert git.workdir == new_workdir
        assert git.workdir == original_workdir

    def test_does_nothing_if_dont_disrupt_config_is_false(self, tmp_git_repo):
        dir = tmp_git_repo(initial_structure={"src": {"hello.txt": "hello git"}})
        git = Git(rootdir=dir)
        original_workdir = git.workdir

        with git.preserve_state(Config.get_default().set_vcs(dont_disrupt_prior_state=False).vcs) as new_workdir:
            assert len(bash.execute(["git", "-C", dir, "worktree", "list"]).output.splitlines()) == 1
            assert new_workdir == original_workdir



import logging
import os
import typing as t
from pathlib import Path

from automyte.utils.filesystem import parse_dir

from ..file import File, OSFile
from ..filters import Filter
from .base import ProjectExplorer

logger = logging.getLogger(__name__)

IGNORE_FILES_LIST_PATTERNS = [
    ".git",
    "node_modules/",
]


class LocalFilesExplorer(ProjectExplorer):
    def __init__(
        self,
        rootdir: str,
        filter_by: Filter | None = None,
        ignore_locations: list[str] = IGNORE_FILES_LIST_PATTERNS,
    ):
        rootdir = str(parse_dir(rootdir))
        self.rootdir = rootdir
        self.filter_by = filter_by
        self._changed_files: list[OSFile] = []
        self.ignore_locations = ignore_locations

    def _all_files(self) -> t.Generator[OSFile, None, None]:
        for root, dirs, files in os.walk(self.rootdir):
            for f in files:
                file = OSFile(fullname=str(Path(root) / f))
                if self.ignore_locations and self._should_ignore_file(file):
                    continue
                else:
                    yield file

    def explore(self) -> t.Generator[OSFile, None, None]:
        for file in self._all_files():
            # Don't filter at all if no filters supplied or actually apply them
            if not self.filter_by or self.filter_by.filter(file):
                yield file

                if file.is_tainted:
                    self._changed_files.append(file)

    def get_rootdir(self) -> str:
        return self.rootdir

    def set_rootdir(self, newdir: str) -> str:
        self.rootdir = newdir
        return newdir

    def flush(self):
        logger.debug("[Explorer %s]: Flushing following files: %s", self.rootdir, self._changed_files)
        for file in self._changed_files:
            file.flush()

    def _should_ignore_file(self, file: OSFile) -> bool:
        relative_to_rootdir_path = file.fullpath.resolve().relative_to(Path(self.rootdir).resolve())
        for pattern in self.ignore_locations:
            if pattern in str(relative_to_rootdir_path):
                return True

        return False



import typing as t
from dataclasses import dataclass

from .vcs import SupportedVCS, VCSConfig

RUN_MODES = t.Literal["run", "amend"]

_ProjectID: t.TypeAlias = str
AutomatonTarget: t.TypeAlias = t.Literal["all", "new", "successful", "failed", "skipped"] | _ProjectID


@dataclass
class Config:
    mode: RUN_MODES
    vcs: VCSConfig
    stop_on_fail: bool = True
    target: AutomatonTarget = "all"

    @classmethod
    def get_default(cls, **kwargs):
        # TODO: Fix error when passing mode='run' in get_default kwargs and that gets multiple values for same arg.
        return cls(
            mode="run",
            stop_on_fail=True,
            vcs=VCSConfig.get_default(),
            **kwargs,
        )

    def set_vcs(self, **kwargs):
        self.vcs = VCSConfig.get_default(**kwargs)
        return self



import csv
from pathlib import Path

import pytest

from automyte.utils.random import random_hash


@pytest.fixture
def tmp_csv_file(tmp_local_project):
    def _tmp_csv_file_factory(contents: list[list], filename: str | None = None, delimiter: str = "|") -> Path:
        filename = filename or random_hash()
        dir = tmp_local_project(structure={filename: ""})
        filepath = Path(dir) / filename

        with open(filepath, "w") as csv_file:
            writer = csv.writer(csv_file, delimiter=delimiter)
            writer.writerows(contents)

        return filepath

    return _tmp_csv_file_factory



from __future__ import annotations

import os
import typing as t
from pathlib import Path

from .base import File


class OSFile(File):
    def __init__(self, fullname: str):
        self._initial_location = fullname
        self._location = fullname

        self._inital_contents: str | None = None
        self._contents: str | None = None

        self._marked_for_delete: bool = False
        self.tainted: bool = False

    @property
    def folder(self) -> str:
        return str(Path(self._location).parent)

    @property
    def name(self) -> str:
        return str(Path(self._location).name)

    @property
    def fullpath(self) -> Path:
        return Path(self._location)

    def read(self) -> "OSFile":
        # If the file was moved before initial read - we need to read from initial location.
        current_location = self._location if self.fullpath.exists() else self._initial_location

        with open(current_location, "r") as physical_file:
            self._inital_contents = physical_file.read()
            self._contents = self._inital_contents

        return self

    def flush(self) -> None:
        if not self.tainted:
            return

        if self._marked_for_delete:
            if self.fullpath.exists():
                self.fullpath.unlink()
            return

        # If file has been moved - self._location will have been changed, so we write to new location.
        with open(self._location, "w") as physical_file:
            physical_file.write(self.get_contents())

        # Cleanup old file after move() call.
        if self._initial_location != self._location:
            Path(self._initial_location).unlink()

    def contains(self, text: str) -> bool:
        return text in self.get_contents()

    def move(self, to: str | None = None, new_name: str | None = None) -> File:
        self._location = str(Path(to or self.folder) / (new_name or self.name))
        self.tainted = True
        return self

    def get_contents(self) -> str:
        if self._contents is None:
            self.read()

        return self._contents or ""

    def edit(self, text: str) -> File:
        self._contents = text
        self.tainted = True
        return self

    def delete(self) -> File:
        self._marked_for_delete = True
        self.tainted = True
        return self

    @property
    def is_tainted(self) -> bool:
        return self.tainted

    def __str__(self):
        return self._location



import typing as t

from automyte import (
    Automaton,
    Config,
    ContainsFilter,
    File,
    LocalFilesExplorer,
    Project,
    RunContext,
    TasksFlow,
    conditionals,
    guards,
)


def lol(ctx: RunContext, file: File | None):
    import re

    if file:
        file.edit(re.sub(r"world", "there", file.get_contents()))


def test_guards_simple(tmp_local_project):
    dir = tmp_local_project(structure={"src": {"hello.txt": "hello world!"}})

    Automaton(
        name="impl1",
        config=Config.get_default().set_vcs(dont_disrupt_prior_state=False),
        projects=[
            Project(
                project_id="test_project",
                explorer=LocalFilesExplorer(rootdir=dir, filter_by=ContainsFilter(contains="hello world")),
            ),
        ],
        tasks=TasksFlow(
            [
                conditionals.RunOn(
                    lol,
                    lol,
                    lol,
                    on=guards.MODE.amend,
                ),
            ]
        ),
    ).run()

    with open(f"{dir}/src/hello.txt", "r") as f:
        assert f.read() == "hello world!"



from datetime import datetime, timedelta
from unittest.mock import patch

from automyte.automaton.run_context import RunContext
from automyte.discovery import OSFile
from automyte.tasks import vcs
from automyte.utils import bash


class TestVCSTaskAdd:
    def test_stages_files_for_git_vcs(self, tmp_git_repo, run_ctx):
        dir = tmp_git_repo(
            initial_structure={"src": {"hello.txt": "hello vcs task"}},
            unstaged_structure={"src": {"hello.txt": "hello add", "new_file.txt": "hello git"}},
        )
        ctx = run_ctx(dir=dir)
        file = OSFile(fullname=f"{dir}/src/hello.txt")

        vcs.add("src")(ctx, file)

        staged_files_diff = bash.execute(["git", "-C", dir, "diff", "--name-only", "--cached"]).output
        assert "src/hello.txt" in staged_files_diff
        assert "src/new_file.txt" in staged_files_diff

    def test_properly_adds_command_flags_for_git_vcs(self, tmp_git_repo, run_ctx):
        dir = tmp_git_repo(
            initial_structure={"src": {"hello.txt": "hello vcs task"}},
            unstaged_structure={"src": {"hello.txt": "hello add"}},
        )
        ctx = run_ctx(dir=dir)
        file = OSFile(fullname=f"{dir}/src/hello.txt")

        vcs.add("src").flags("--dry-run")(ctx, file)

        staged_files_diff = bash.execute(["git", "-C", dir, "diff", "--name-only", "--cached"]).output
        assert "src/hello.txt" not in staged_files_diff


class TestVCSTaskCommit:
    def test_creates_commit_for_git_vcs(self, tmp_git_repo, run_ctx):
        dir = tmp_git_repo(
            initial_structure={"src": {"hello.txt": "hello vcs task"}},
            unstaged_structure={"src": {"hello.txt": "hello new commit"}},
        )
        ctx = run_ctx(dir=dir)
        file = OSFile(fullname=f"{dir}/src/hello.txt")
        vcs.add(".")(ctx, file)

        vcs.commit("commit #2")(ctx, file)

        commits_list = bash.execute(["git", "-C", dir, "log", "--oneline"]).output.split("\n")
        assert "commit #2" in commits_list[0]

    def test_doesnt_alter_any_other_commits_for_git_vcs(self, tmp_git_repo, run_ctx):
        dir = tmp_git_repo(
            initial_structure={"src": {"hello.txt": "hello vcs task"}},
            unstaged_structure={"src": {"hello.txt": "hello new commit"}},
        )
        ctx = run_ctx(dir=dir)
        file = OSFile(fullname=f"{dir}/src/hello.txt")
        prev_commits = bash.execute(["git", "-C", dir, "log", "--oneline"]).output.split("\n")
        vcs.add(".")(ctx, file)

        vcs.commit("some message")(ctx, file)

        commits_list = bash.execute(["git", "-C", dir, "log", "--oneline"]).output.split("\n")
        assert len(commits_list) == len(prev_commits) + 1
        assert commits_list[1] == prev_commits[0]

    def test_properly_adds_command_flags_for_git_vcs(self, tmp_git_repo, run_ctx):
        dir = tmp_git_repo(
            initial_structure={"src": {"hello.txt": "hello vcs task"}},
            unstaged_structure={"src": {"hello.txt": "hello new commit"}},
        )
        ctx = run_ctx(dir=dir)
        file = OSFile(fullname=f"{dir}/src/hello.txt")

        vcs.commit("commit #2").flags("--include", "src")(ctx, file)

        changed_files = bash.execute(["git", "-C", dir, "log", "-1", "--pretty=format:", "--name-only"]).output
        assert "hello.txt" in changed_files


class TestVCSTaskPush:
    def test_calls_correct_cmd_for_git_vcs(self, tmp_git_repo, run_ctx):
        dir = tmp_git_repo(
            initial_structure={"src": {"hello.txt": "hello vcs task"}},
            unstaged_structure={"src": {"hello.txt": "hello new commit"}},
        )
        ctx = run_ctx(dir=dir)
        file = OSFile(fullname=f"{dir}/src/hello.txt")

        with patch("automyte.utils.bash.execute") as mock_execute:
            vcs.push(to="automyte123123123")(ctx, file)
            mock_execute.assert_called_once_with(["git", "-C", dir, "push", "origin", "automyte123123123"])

    def test_properly_adds_command_flags_for_git_vcs(self, tmp_git_repo, run_ctx):
        dir = tmp_git_repo(
            initial_structure={"src": {"hello.txt": "hello vcs task"}},
            unstaged_structure={"src": {"hello.txt": "hello new commit"}},
        )
        ctx = run_ctx(dir=dir)
        file = OSFile(fullname=f"{dir}/src/hello.txt")

        with patch("automyte.utils.bash.execute") as mock_execute:
            vcs.push(to="automyte123123123").flags("--force-with-lease")(ctx, file)
            mock_execute.assert_called_once_with(
                ["git", "-C", dir, "push", "--force-with-lease", "origin", "automyte123123123"]
            )



import pytest
import contextlib

from automyte.automaton.automaton import Automaton, Config
from automyte.automaton.types import FileTask
from automyte.discovery import ProjectExplorer, OSFile
from automyte.project import Project
from automyte.history.in_memory import InMemoryHistory
from automyte.vcs import VCS


class DummyExplorer(ProjectExplorer):
    def __init__(self) -> None:
        self.rootdir = 'smth'

    def get_rootdir(self) -> str:
        return self.rootdir

    def set_rootdir(self, newdir: str):
        self.rootdir = newdir
        return self.rootdir

    def flush(self):
        self.flushed = True

    def explore(self):
        yield OSFile(fullname=f"{self.rootdir}/hello.txt")


class DummyVCS(VCS):
    def run(self, *args):
        return

    @contextlib.contextmanager
    def preserve_state(self, config):
        self.preserve_state_called = True
        yield "newdir"


@pytest.fixture
def dummy_explorer():
    def _explorer_factory():
        return DummyExplorer()

    return _explorer_factory

@pytest.fixture
def dummy_vcs():
    def _vcs_factory():
        return DummyVCS()

    return _vcs_factory

@pytest.fixture
def dummy_automaton():

    def _dummy_automaton_factory(
        name: str = 'auto',
        projects: list[Project] | None = None,
        tasks: list[FileTask] | None = None
    ):
        return Automaton(
            name=name,
            config=Config.get_default().set_vcs(dont_disrupt_prior_state=False),
            history=InMemoryHistory(),
            projects=projects or [],
            tasks=tasks or [],
        )

    return _dummy_automaton_factory



from automyte.automaton.run_context import RunContext
from automyte.automaton.types import TaskReturn
from automyte.discovery.file.os_file import OSFile
from automyte.tasks import conditionals as cd


class TestRunIf:
    def test_skips_execution_if_condition_is_false(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("hello conditional")
        ctx: RunContext = run_ctx(dir=file.folder)

        has_been_called = []  # Will just check for faulty.
        some_task = lambda ctx, file: has_been_called.append(1)

        cd.RunIf(some_task, condition=False)(ctx, file)

        assert not has_been_called

    def test_runs_tasks_if_condition_is_true(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("hello conditional")
        ctx: RunContext = run_ctx(dir=file.folder)

        has_been_called = []  # Will just check for faulty.
        some_task = lambda ctx, file: has_been_called.append(1)

        cd.RunIf(some_task, condition=True)(ctx, file)

        assert has_been_called

    def test_properly_runs_for_multiple_tasks(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("hello conditional")
        ctx: RunContext = run_ctx(dir=file.folder)

        has_been_called = []  # Will just check for faulty.
        some_task = lambda ctx, file: has_been_called.append(1)

        cd.RunIf(some_task, some_task, condition=True)(ctx, file)

        assert len(has_been_called) == 2

    def test_stops_execution_if_one_of_the_tasks_fail(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("hello conditional")
        ctx: RunContext = run_ctx(dir=file.folder)

        has_been_called = []  # Will just check for faulty.
        failure_of_a_task = lambda ctx, file: TaskReturn(instruction="skip")
        some_task = lambda ctx, file: has_been_called.append(1)

        cd.RunIf(failure_of_a_task, some_task, condition=True)(ctx, file)
        assert not has_been_called


class TestRunOn:
    def test_skips_execution_if_condition_is_false(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("hello conditional")
        ctx: RunContext = run_ctx(dir=file.folder)

        has_been_called = []  # Will just check for faulty.
        some_task = lambda ctx, file: has_been_called.append(1)
        checker = lambda ctx, file: False

        cd.RunOn(some_task, on=checker)(ctx, file)

        assert not has_been_called

    def test_runs_tasks_if_condition_is_true(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("hello conditional")
        ctx: RunContext = run_ctx(dir=file.folder)

        has_been_called = []  # Will just check for faulty.
        some_task = lambda ctx, file: has_been_called.append(1)
        checker = lambda ctx, file: True

        cd.RunOn(some_task, on=checker)(ctx, file)

        assert has_been_called

    def test_properly_runs_for_multiple_tasks(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("hello conditional")
        ctx: RunContext = run_ctx(dir=file.folder)

        has_been_called = []  # Will just check for faulty.
        some_task = lambda ctx, file: has_been_called.append(1)
        checker = lambda ctx, file: True

        cd.RunOn(some_task, some_task, on=checker)(ctx, file)

        assert len(has_been_called) == 2



import re

from automyte import (
    Automaton,
    AutomatonRunResult,
    Config,
    ContainsFilter,
    File,
    InMemoryHistory,
    LocalFilesExplorer,
    Project,
    RunContext,
    TaskReturn,
    TasksFlow,
    guards,
)


def replace(ctx: RunContext, file: File):
    file.edit(re.sub(r"world", "there", file.get_contents()))
    return file.contains("oo")


def second_task(ctx: RunContext, file: File):
    if ctx.previous_return.value is True:
        file.edit(re.sub(r"good", "bad", file.get_contents()))


def test_simple_return_with_one_file(tmp_local_project):
    dir = tmp_local_project(structure={"src": {"hello.txt": "hello world!"}})

    Automaton(
        name="impl1",
        config=Config.get_default().set_vcs(dont_disrupt_prior_state=False),
        projects=[
            Project(
                project_id="test_project",
                explorer=LocalFilesExplorer(rootdir=dir, filter_by=ContainsFilter(contains="hello world")),
            )
        ],
        tasks=TasksFlow([replace, second_task]),
    ).run()

    with open(f"{dir}/src/hello.txt", "r") as f:
        assert f.read() == "hello there!"


def test_simple_return_with_multiple_files(tmp_local_project):
    dir = tmp_local_project(structure={"src": {"hello.txt": "hello world!", "bye.txt": "good bye!"}})

    Automaton(
        name="impl1",
        config=Config.get_default().set_vcs(dont_disrupt_prior_state=False),
        projects=[
            Project(
                project_id="test_project",
                explorer=LocalFilesExplorer(rootdir=dir, filter_by=ContainsFilter(contains="o")),
            )
        ],
        tasks=TasksFlow([replace, second_task]),
    ).run()

    with open(f"{dir}/src/bye.txt", "r") as f:
        assert f.read() == "bad bye!"



from automyte import (
    Automaton,
    AutomatonRunResult,
    Config,
    ContainsFilter,
    File,
    InMemoryHistory,
    LocalFilesExplorer,
    Project,
    RunContext,
    TasksFlow,
    guards,
)


def test_updates_history_to_success(tmp_local_project):
    rootdir1 = tmp_local_project(structure={"src": {"hello.txt": "hello there"}})
    rootdir2 = tmp_local_project(structure={"src": {"hello.txt": "hello there"}})

    history = InMemoryHistory()
    filters = ContainsFilter(contains="hello")
    automaton = Automaton(
        name="hello",
        config=Config.get_default().set_vcs(dont_disrupt_prior_state=False),
        projects=[
            Project(explorer=LocalFilesExplorer(rootdir=rootdir1, filter_by=filters), project_id="proj1"),
            Project(explorer=LocalFilesExplorer(rootdir=rootdir2, filter_by=filters), project_id="proj2"),
        ],
        tasks=TasksFlow([lambda ctx, file: None]),
        history=history,
    )

    automaton.run()

    assert history.get_status("hello", "proj1") == AutomatonRunResult(status="success")
    assert history.get_status("hello", "proj2") == AutomatonRunResult(status="success")


def test_updates_to_fail(tmp_local_project):
    rootdir1 = tmp_local_project(structure={"src": {"hello.txt": "hello there"}})
    rootdir2 = tmp_local_project(structure={"src": {"hello.txt": "hello there"}})

    history = InMemoryHistory()
    filters = ContainsFilter(contains="hello")
    automaton = Automaton(
        name="hello",
        config=Config.get_default().set_vcs(dont_disrupt_prior_state=False),
        projects=[
            Project(explorer=LocalFilesExplorer(rootdir=rootdir1, filter_by=filters), project_id="proj1"),
            Project(explorer=LocalFilesExplorer(rootdir=rootdir2, filter_by=filters), project_id="proj2"),
        ],
        tasks=TasksFlow([lambda ctx, file: exec('raise ValueError("forced")')]),
        history=history,
    )

    automaton.run()

    assert history.get_status("hello", "proj1") == AutomatonRunResult(status="fail", error="forced")
    # Config.stop_on_fail = True by default, so we should never reach proj2, so it should be in "not_run" status.
    assert history.get_status("hello", "proj2") == AutomatonRunResult(status="new")



from collections import defaultdict

from .base import History
from .types import AutomatonRunResult


class InMemoryHistory(History):
    def __init__(self) -> None:
        self.data: dict[str, AutomatonRunResult] = defaultdict(lambda: AutomatonRunResult(status="new"))

    def get_status(self, automaton_name: str, project_id: str) -> AutomatonRunResult:
        return self.data[project_id]

    def set_status(self, automaton_name: str, project_id: str, status: AutomatonRunResult):
        self.data[project_id] = status

    def read(self, automaton_name: str):
        return self.data



import typing as t
from dataclasses import dataclass

SupportedVCS: t.TypeAlias = t.Literal["git"]


@dataclass
class VCSConfig:
    default_vcs: SupportedVCS
    main_branch: str = "master"
    work_branch: str = "automate"
    dont_disrupt_prior_state: bool = True

    @classmethod
    def get_default(cls, **kwargs):
        kwargs.setdefault("default_vcs", "git")
        kwargs.setdefault("main_branch", "master")
        kwargs.setdefault("work_branch", "automate")
        kwargs.setdefault("dont_disrupt_prior_state", True)

        return cls(**kwargs)



import os
from pathlib import Path

from automyte import OSFile


class TestOSFileGetContents:
    def test_returns_file_contents_from_the_disk(self, tmp_os_file):
        file: OSFile = tmp_os_file("osfile")
        assert file.get_contents() == "osfile"

    def test_second_call_doesnt_reread_from_the_dist(self, tmp_os_file):
        file: OSFile = tmp_os_file("osfile")
        file.get_contents()

        new_data = "new_data_shouldnt_equal_in_second_get_contents_call"
        with open(file.fullpath, "w") as disk_file:
            disk_file.write(new_data)

        assert file.get_contents() != new_data


class TestOSFileContains:
    def test_returns_true_if_text_is_in_file(self, tmp_os_file):
        file: OSFile = tmp_os_file("osfile")
        assert file.contains("os")

    def test_returns_false_if_text_is_not_in_file(self, tmp_os_file):
        file: OSFile = tmp_os_file("osfile")
        assert not file.contains("not here")


class TestOSFileEdit:
    def test_overrides_contents(self, tmp_os_file):
        file: OSFile = tmp_os_file("revision 1")
        file.edit(text="rev2")
        assert file.get_contents() == "rev2"

    def test_doesnt_override_disk_contents_immediately(self, tmp_os_file):
        file: OSFile = tmp_os_file("revision 1")
        file.edit(text="rev2")

        with open(Path(file.folder) / file.name, "r") as file_on_disk:
            assert file_on_disk.read() == "revision 1"

    def test_marks_file_as_tainted(self, tmp_os_file):
        file: OSFile = tmp_os_file("revision 1")
        file.edit(text="rev2")

        assert file.is_tainted


class TestOSFileDelete:
    def test_doesnt_delete_disk_file_immediately(self, tmp_os_file):
        file: OSFile = tmp_os_file("revision 1")
        file.delete()

        assert os.path.exists(file.fullpath)

    def test_marks_file_as_tainted(self, tmp_os_file):
        file: OSFile = tmp_os_file("revision 1")
        file.delete()

        assert file.is_tainted

    def test_doesnt_modify_file_contents(self, tmp_os_file):
        file: OSFile = tmp_os_file("revision 1")
        file.delete()

        assert file.get_contents() == "revision 1"


class TestOSFileMove:
    def test_doesnt_move_disk_file_immediately(self, tmp_os_file):
        file: OSFile = tmp_os_file("revision 1")
        old_path = file.fullpath
        file.move(new_name="newname")

        assert not os.path.exists(file.fullpath)
        assert os.path.exists(old_path)

    def test_marks_file_as_tainted(self, tmp_os_file):
        file: OSFile = tmp_os_file("revision 1")
        file.move(new_name="newname")

        assert file.is_tainted

    def test_doesnt_modify_file_contents(self, tmp_os_file):
        file: OSFile = tmp_os_file("revision 1")
        file.move(new_name="newname")

        assert file.get_contents() == "revision 1"

    def test_both_arguments_update_fullpath(self, tmp_local_project):
        dir = tmp_local_project({"src": {"subdir1": {"oldname.txt": "revision 1"}, "subdir2": {}}})
        file = OSFile(fullname=f"{dir}/subdir1/oldname.txt")

        file.move(to=f"{dir}/subdir2", new_name="newname.txt")

        assert file.fullpath == Path(dir) / "subdir2" / "newname.txt"


class TestOSFileFlush:
    def test_flush_overrides_disk_contents_after_edit(self, tmp_os_file):
        file: OSFile = tmp_os_file("revision 1")
        file.edit(text="rev2").flush()

        with open(Path(file.folder) / file.name, "r") as file_on_disk:
            assert file_on_disk.read() == "rev2"

    def test_flush_deletes_file_after_delete(self, tmp_os_file):
        file: OSFile = tmp_os_file("revision 1")
        file.delete().flush()

        assert not os.path.exists(file.fullpath)

    def test_flush_moves_file_to_new_location_after_move(self, tmp_os_file):
        file: OSFile = tmp_os_file("revision 1")
        file.move(new_name="newname").flush()

        assert os.path.exists(file.fullpath)

    def test_flush_removes_old_file_after_move(self, tmp_os_file):
        file: OSFile = tmp_os_file("revision 1")
        old_path = file.fullpath
        file.move(new_name="newname").flush()

        assert not os.path.exists(old_path)


class TestOSFileProperties:
    def test_extension_is_included_in_the_name(self, tmp_os_file):
        file: OSFile = tmp_os_file("whatever", filename="my_file.txt")
        assert file.name == "my_file.txt"



import re
from pathlib import Path

from .base import File, Filter


class PathFilter(Filter):
    """Filter files by their filename or folder.

    `filename` is treated as regexp so it is possible to provide smth like `filename=".*.py"`
        to search for file extension for example.
    `folder` search is performed by just checking if given folder is present in file path

    If both params are passed - they are treated as "and" filter.
    """

    def __init__(
        self,
        filename: str | None = None,
        folder: str | Path | None = None,
    ) -> None:
        self.name = filename

        if isinstance(folder, Path):
            self.folder = folder
        elif isinstance(folder, str):
            self.folder = Path(folder)
        else:
            self.folder = None

    def filter(self, file: File) -> bool:
        if self.name:
            if not re.search(self.name, file.name):
                return False

        if self.folder:
            if str(self.folder) not in file.folder:
                return False

        return True



from unittest.mock import patch

from automyte import OSFile, RunContext
from automyte.tasks.utils import Breakpoint


class TestBreakpoint:
    def test_continues_by_default_input(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("whatever")
        ctx: RunContext = run_ctx(file.folder)

        with patch("automyte.tasks.utils.breakpoint.Breakpoint._get_input", return_value=""):
            result = Breakpoint()(ctx, file)

        assert result is None

    def test_option_d_prints_project_dir_and_waits_in_loop(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("whatever")
        ctx: RunContext = run_ctx(file.folder)

        with (
            patch("automyte.tasks.utils.breakpoint.Breakpoint._get_input", side_effect=["p", "c"]) as mock_input,
            patch("builtins.print") as mock_print,
        ):
            Breakpoint()(ctx, file)

        assert mock_input.call_count == 2
        mock_print.assert_called_once_with(file.folder)

    def test_option_a_flushes_files_and_waits_in_loop(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("whatever")
        ctx: RunContext = run_ctx(file.folder)

        with (
            patch("automyte.tasks.utils.breakpoint.Breakpoint._get_input", side_effect=["a", "c"]) as mock_input,
            patch("automyte.project.project.Project.apply_changes") as apply_mock,
        ):
            Breakpoint()(ctx, file)

        assert mock_input.call_count == 2
        apply_mock.assert_called_once()



import typing as t

from automyte.automaton import BaseTask, RunContext
from automyte.automaton.flow import execute_tasks_sequence
from automyte.automaton.types import TaskReturn
from automyte.discovery import File


class RunIf:
    def __init__(self, *tasks: BaseTask, condition: bool):
        self.condition = condition
        self.tasks = list(tasks)

    def __call__(self, ctx: RunContext, file: File | None):
        if not self.condition:
            return TaskReturn(status="skipped")

        return execute_tasks_sequence(tasks=self.tasks, ctx=ctx, file=file)


class RunOn:
    def __init__(self, *tasks: BaseTask, on: t.Callable[[RunContext, File | None], bool]) -> None:
        self.validator = on
        self.tasks = list(tasks)

    def __call__(self, ctx: RunContext, file: File | None):
        if not self.validator(ctx, file):
            return TaskReturn(status="skipped")

        return execute_tasks_sequence(tasks=self.tasks, ctx=ctx, file=file)



from .explorers import LocalFilesExplorer, ProjectExplorer
from .file import File, OSFile
from .filters import ContainsFilter, Filter, PathFilter

__all__ = [
    "File",
    "Filter",
    "LocalFilesExplorer",
    "ContainsFilter",
    "OSFile",
    "ProjectExplorer",
    "PathFilter",
]



from __future__ import annotations

from pathlib import Path

from automyte.automaton.run_context import RunContext
from automyte.discovery import File


class VCSTask:
    def __init__(self):
        self._flags: list[str] = []

    def __call__(self, ctx: RunContext, file: File | None = None):
        raise NotImplementedError

    def flags(self, *args: str) -> "VCSTask":
        self._flags.extend(list(args))
        return self


class add(VCSTask):
    def __init__(self, paths: str | Path | list[str | Path]):
        super().__init__()
        if isinstance(paths, str) or isinstance(paths, Path):
            self.paths = [paths]
        else:
            self.paths = paths

    def __call__(self, ctx: RunContext, file: File | None = None):
        ctx.vcs.run("add", "--", *[str(p) for p in self.paths], *self._flags)


class commit(VCSTask):
    def __init__(self, msg: str):
        super().__init__()
        self.msg = msg

    def __call__(self, ctx: RunContext, file: File | None = None):
        ctx.vcs.run("commit", "-m", self.msg, *self._flags)


class push(VCSTask):
    def __init__(self, to: str, remote: str = "origin"):
        super().__init__()
        self.to = to
        self.remote = remote

    def __call__(self, ctx: RunContext, file: File | None = None):
        ctx.vcs.run("push", *self._flags, self.remote, self.to)



from __future__ import annotations

import contextlib
import typing as t
from pathlib import Path
from urllib.parse import urlparse

from automyte.config import Config
from automyte.discovery import LocalFilesExplorer, ProjectExplorer
from automyte.utils.filesystem import parse_dir
from automyte.utils.random import random_hash
from automyte.vcs import VCS, Git


class Project:
    def __init__(
        self,
        project_id: str,
        rootdir: str | None = None,
        explorer: ProjectExplorer | None = None,
        vcs: VCS | None = None,
    ):
        if not rootdir and not explorer:
            raise ValueError("Need to supply at least one of: rootdir | explorer")
        if rootdir:
            rootdir = str(parse_dir(rootdir))

        self.project_id = project_id
        self.explorer = explorer or LocalFilesExplorer(rootdir=t.cast(str, rootdir))
        self.rootdir = rootdir or self.explorer.get_rootdir()
        self.vcs = vcs or Git(rootdir=self.rootdir)

    @contextlib.contextmanager
    def in_working_state(self, config: Config):
        """Hook for doing any initial project setup and cleanup after the work is done.

        Real use case for now - adjusting rootdir to point to worktree dir if dont_disrupt_prior_state = True for Git.
        """
        # Setup phase:
        original_rootdir = self.rootdir
        with self.vcs.preserve_state(config=config.vcs) as current_project_dir:
            self.rootdir = str(current_project_dir)
            self.explorer.set_rootdir(newdir=str(current_project_dir))

            yield

        self.rootdir = original_rootdir
        self.explorer.set_rootdir(newdir=self.rootdir)

    def apply_changes(self):
        self.explorer.flush()

    @classmethod
    def from_uri(cls, uri: str) -> "Project":
        parsed_uri = urlparse(uri)
        # TODO: Add processing for file:///... schema
        if not parsed_uri.scheme:
            path = parse_dir(uri)
            project_id = f"{random_hash(str(path.parent))}_{path.name}"
            return cls(rootdir=str(path), project_id=project_id)

        else:
            raise NotImplementedError("Cloud path or 'file:///...' is not available yet.")



from automyte import Automaton, AutomatonRunResult, InMemoryHistory, OSFile, Project, RunContext, guards
from automyte.automaton.types import TaskReturn


class TestModeGuards:
    def test_run(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("hello conditional")
        ctx: RunContext = run_ctx(dir=file.folder)
        ctx.config.mode = "run"

        assert guards.MODE.run(ctx, file)
        assert not guards.MODE.amend(ctx, file)

    def test_amend(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("hello conditional")
        ctx: RunContext = run_ctx(dir=file.folder)
        ctx.config.mode = "amend"

        assert not guards.MODE.run(ctx, file)
        assert guards.MODE.amend(ctx, file)


class TestHistoryGuards:
    def test_failed(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("smth")
        ctx: RunContext = run_ctx(file.folder)
        ctx.previous_status = AutomatonRunResult("fail")

        assert guards.HISTORY.failed(ctx, file)

    def test_new(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("smth")
        ctx: RunContext = run_ctx(file.folder)
        ctx.previous_status = AutomatonRunResult("new")

        assert guards.HISTORY.new(ctx, file)

    def test_skipped(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("smth")
        ctx: RunContext = run_ctx(file.folder)
        ctx.previous_status = AutomatonRunResult("skipped")

        assert guards.HISTORY.skipped(ctx, file)

    def test_succeeded(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("smth")
        ctx: RunContext = run_ctx(file.folder)
        ctx.previous_status = AutomatonRunResult("success")

        assert guards.HISTORY.succeeded(ctx, file)


class TestPreviousTaskGuards:
    def test_is_success(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("whatever")
        ctx: RunContext = run_ctx(file.folder)
        ctx.save_task_result(result=TaskReturn(status="processed"), file=file)

        assert guards.PREVIOUS_TASK.is_success(ctx, file)

    def test_was_skipped(self, run_ctx, tmp_os_file):
        file: OSFile = tmp_os_file("whatever")
        ctx: RunContext = run_ctx(file.folder)
        ctx.save_task_result(result=TaskReturn(status="skipped"), file=file)

        assert guards.PREVIOUS_TASK.was_skipped(ctx, file)



import csv
import logging
import os
import typing as t
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path

from .base import History
from .types import AutomatonRunResult, ProjectID, RunStatus

logger = logging.getLogger(__name__)


class InFileHistory(History):
    """Store automatons runs for projects in a csv file.

    If history filename is reused between automatons - data will just be appended and separated for automatons,
        not overriden or lost.
    File extension doesn't matter, internally, file structure is csv.
    Users are free to manipulate that file manually, for manual overrides.
    """

    def __init__(self, filename: str | Path | None = None) -> None:
        """Setup new instance of in-file history.

        If filename is a directory - will create a new "automyte_history.csv" file there;
        If filename is a full path to a file - will either override/read contents or create new file with that name,
            as long as the parent folder of that file exists.

        Can set filename to either None, "./", "current", "local" to create history file in the script launch dir.
        """
        self.filepath = self._set_storage_path(filename)
        self.file_io = _HistoryFileIO(filepath=self.filepath)

    def get_status(self, automaton_name: str, project_id: str) -> AutomatonRunResult:
        return self.file_io.get_automaton_history(automaton_name=automaton_name).get(
            project_id, AutomatonRunResult("new", error=None)
        )

    def set_status(self, automaton_name: str, project_id: str, status: AutomatonRunResult):
        self.file_io.save_project_run(automaton_name=automaton_name, project_id=project_id, run_result=status)

    def read(self, automaton_name: str):
        if not self.filepath.parent.exists():
            logger.error("[History %s]: Folder %s doesn't exist.", automaton_name, self.filepath)
            raise ValueError(f"Path {self.filepath} does not exist")

        return self.file_io.get_automaton_history(automaton_name=automaton_name)

    def _set_storage_path(self, filepath: str | Path | None):
        result = Path(os.getcwd())

        if isinstance(filepath, str):
            if filepath == "." or filepath == "./" or filepath == "current" or filepath == "local":
                pass
            else:
                result = Path(filepath)

        elif isinstance(filepath, Path):
            result = filepath

        if result.is_dir():
            result = result / "automyte_history.csv"
        return result


@dataclass
class _AutomatonHistoryInstance:
    automaton_name: str
    projects: dict[ProjectID, AutomatonRunResult]


class _HistoryFileIO:
    def __init__(self, filepath: Path) -> None:
        self.filepath = filepath
        self.separator = "|"

    def save_project_run(self, automaton_name: str, project_id: ProjectID, run_result: AutomatonRunResult):
        history_instances = {
            automaton: _AutomatonHistoryInstance(automaton_name=automaton, projects=projects)
            for automaton, projects in self._parse_file().items()
        }

        if not history_instances.get(automaton_name):
            history_instances[automaton_name] = _AutomatonHistoryInstance(
                automaton_name=automaton_name, projects={project_id: run_result}
            )
        else:
            history_instances[automaton_name].projects[project_id] = run_result

        self._update_history_file(all_automatons_histories=list(history_instances.values()))

    def get_automaton_history(self, automaton_name: str) -> dict[ProjectID, AutomatonRunResult]:
        data = self._parse_file()
        return data.get(automaton_name, {})

    def _parse_file(self) -> dict[str, dict[ProjectID, AutomatonRunResult]]:
        history_instances: t.DefaultDict[str, dict[ProjectID, AutomatonRunResult]] = defaultdict(dict)

        if not self.filepath.exists():
            self.filepath.touch()
            return {}

        with open(self.filepath, "r") as csv_file:
            reader = csv.reader(csv_file, delimiter=self.separator)
            # Skipping header, as we don't care.
            next(reader)

            for automaton_name, project_id, status, error in reader:
                if status not in ("fail", "success", "skipped", "running", "new"):
                    logger.error(
                        "[History %s]: Failed to parse history file - wrong project status: %s.",
                        automaton_name,
                        status,
                    )
                    raise ValueError(f"Incorrect project status in history file: {status}")
                history_instances[automaton_name][project_id] = AutomatonRunResult(status=status, error=error or None)

        return history_instances

    def _update_history_file(self, all_automatons_histories: list[_AutomatonHistoryInstance]):
        with open(self.filepath, "w") as csv_file:
            writer = csv.DictWriter(
                csv_file, delimiter=self.separator, fieldnames=["automaton", "project", "status", "error"]
            )
            writer.writeheader()

            for automaton_history in all_automatons_histories:
                writer.writerows(
                    [
                        {
                            "automaton": automaton_history.automaton_name,
                            "project": project_id,
                            "status": run_result.status,
                            "error": run_result.error,
                        }
                        for project_id, run_result in automaton_history.projects.items()
                    ]
                )



import enum
import os
import sys
import typing as t


@t.final
class ExitCode(enum.IntEnum):
    OK = 0
    MIGRATION_FAILED = 1


def main() -> ExitCode:
    return ExitCode.OK


def console_main() -> int:
    """The CLI entry point of automyte.

    This function is not meant for programmable use; use `main()` instead.
    """
    # https://docs.python.org/3/library/signal.html#note-on-sigpipe
    try:
        code = main()
        sys.stdout.flush()
    except BrokenPipeError:
        # Python flushes standard streams on exit; redirect remaining output
        # to devnull to avoid another BrokenPipeError at shutdown
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        return 1  # Python exits with error code 1 on EPIPE
    else:
        return code



from automyte.discovery import OSFile
from automyte.discovery.filters import PathFilter


class TestPathFilter:
    def test_correct_filename_match(self, tmp_os_file):
        file: OSFile = tmp_os_file("", filename="hello.txt")

        assert PathFilter(filename="hello.txt").filter(file=file)
        assert not PathFilter("hello.py").filter(file=file)

    def test_filename_is_searched_through_regexp(self, tmp_os_file):
        file: OSFile = tmp_os_file("", filename="hello.txt")

        assert PathFilter(filename=r".*llo.*").filter(file=file)

    def test_filename_allows_searching_by_extension(self, tmp_os_file):
        file: OSFile = tmp_os_file("", filename="hello.txt")

        assert PathFilter(filename=r".*.txt").filter(file=file)

    def test_correct_folder_match(self, tmp_os_file):
        file: OSFile = tmp_os_file("", filename="hello.txt")
        other_file: OSFile = tmp_os_file("")

        assert PathFilter(folder=file.folder).filter(file=file)
        assert not PathFilter(folder=other_file.folder).filter(file=file)

    def test_partial_folder_match(self, tmp_local_project):
        dir = tmp_local_project({"src": {"nested": {"subdir": {"hello.txt": ""}}}})
        file = OSFile(fullname=f"{dir}/nested/subdir/hello.txt")

        assert PathFilter(folder="nested/subdir").filter(file=file)

    def test_if_both_name_and_folder_provided_then_match_together(self, tmp_local_project):
        dir = tmp_local_project({"src": {"nested": {"subdir": {"hello.txt": "", "bye.py": ""}}}})
        file = OSFile(fullname=f"{dir}/nested/subdir/bye.py")
        wrong_file = OSFile(fullname=f"{dir}/nested/subdir/hello.txt")

        assert PathFilter(folder="subdir", filename=r"bye.py").filter(file=file)
        assert not PathFilter(folder="subdir", filename=r"bye.py").filter(file=wrong_file)



import pytest

from automyte import (
    Automaton,
    AutomatonRunResult,
    Config,
    ContainsFilter,
    File,
    InMemoryHistory,
    LocalFilesExplorer,
    Project,
    RunContext,
    TaskReturn,
    TasksFlow,
    guards,
)


def test_targetting_by_target_id(tmp_local_project):
    rootdir1 = tmp_local_project(structure={"src": {"hello.txt": "hello there"}})
    rootdir2 = tmp_local_project(structure={"src": {"hello.txt": "hello there"}})

    history = InMemoryHistory()
    history.set_status("hello", "proj1", AutomatonRunResult(status="fail"))
    history.set_status("hello", "proj2", AutomatonRunResult(status="new"))
    filters = ContainsFilter(contains="hello")

    config = Config.get_default(target="proj1").set_vcs(dont_disrupt_prior_state=False)

    automaton = Automaton(
        name="hello",
        config=config,
        projects=[
            Project(explorer=LocalFilesExplorer(rootdir=rootdir1, filter_by=filters), project_id="proj1"),
            Project(explorer=LocalFilesExplorer(rootdir=rootdir2, filter_by=filters), project_id="proj2"),
        ],
        tasks=TasksFlow([lambda ctx, file: None]),
        history=history,
    )

    automaton.run()

    # Ran for proj1 due to target_id
    assert history.get_status("hello", "proj1") == AutomatonRunResult(status="success")
    assert history.get_status("hello", "proj2") == AutomatonRunResult(status="new")  # Never ran for proj2


@pytest.mark.parametrize(
    "initial_status, target_status",
    [
        ("fail", "failed"),
        ("skipped", "skipped"),
    ],
)
def test_targetting_by_status(tmp_local_project, initial_status, target_status):
    rootdir1 = tmp_local_project(structure={"src": {"hello.txt": "hello there"}})
    rootdir2 = tmp_local_project(structure={"src": {"hello.txt": "hello there"}})

    history = InMemoryHistory()
    history.set_status("hello", "proj1", AutomatonRunResult(status=initial_status))
    history.set_status("hello", "proj2", AutomatonRunResult(status="new"))
    filters = ContainsFilter(contains="hello")

    config = Config.get_default(target=target_status).set_vcs(dont_disrupt_prior_state=False)

    automaton = Automaton(
        name="hello",
        config=config,
        projects=[
            Project(explorer=LocalFilesExplorer(rootdir=rootdir1, filter_by=filters), project_id="proj1"),
            Project(explorer=LocalFilesExplorer(rootdir=rootdir2, filter_by=filters), project_id="proj2"),
        ],
        tasks=TasksFlow([lambda ctx, file: None]),
        history=history,
    )

    automaton.run()

    # Ran for proj1 due to target='failed'
    assert history.get_status("hello", "proj1") == AutomatonRunResult(status="success")
    assert history.get_status("hello", "proj2") == AutomatonRunResult(status="new")  # Never ran for proj2



from .automaton import Automaton
from .flow import TasksFlow
from .run_context import RunContext
from .types import BaseTask, FileTask, InstructionForAutomaton, TaskReturn

__all__ = [
    "Automaton",
    "InstructionForAutomaton",
    "RunContext",
    "FileTask",
    "BaseTask",
    "TaskReturn",
    "TasksFlow",
]



import re

from automyte import Automaton, File, Project, RunContext
from automyte.tasks import fs, vcs
from automyte.utils import bash

file_contents = """
some_var = True
if some_var is True:
    do_stuff()
"""

expected_new_contents = """
some_var = True
if some_var:
    do_stuff()
"""


def test_simplest_setup(tmp_git_repo):
    def remove_redundant_is_true_comparisons(ctx: RunContext, file: File):
        file.edit(re.sub(r"if (\w+) is True:", r"if \1:", file.get_contents()))

    directory = tmp_git_repo({"src": {"some_python_file.py": file_contents}})

    Automaton(
        name="simplest_case",
        projects=[directory],
        tasks=[
            remove_redundant_is_true_comparisons,
            fs.flush(),
            vcs.add("src"),
            vcs.commit("Remove unnecessary 'is True' comparisons"),
        ],
    ).run()

    bash.execute(["git", "-C", directory, "checkout", "automate"])
    with open(f"{directory}/src/some_python_file.py", "r") as proj1_file:
        assert proj1_file.read() == expected_new_contents



from automyte.automaton import *
from automyte.config import *
from automyte.discovery import *
from automyte.history import *
from automyte.project import *
from automyte.tasks import *
from automyte.vcs import *

# TODO: Maybe define actual exports in __all__ ?



class _ModeGuards:
    run = lambda ctx, file: ctx.config.mode == "run"
    amend = lambda ctx, file: ctx.config.mode == "amend"


class _HistoryGuards:
    failed = lambda ctx, file: ctx.previous_status.status == "fail"
    skipped = lambda ctx, file: ctx.previous_status.status == "skipped"
    succeeded = lambda ctx, file: ctx.previous_status.status == "success"
    new = lambda ctx, file: ctx.previous_status.status == "new"


class _PreviousTaskGuards:
    is_success = lambda ctx, file: ctx.previous_return.value is None or ctx.previous_return.status == "processed"
    was_skipped = lambda ctx, file: ctx.previous_return.value is None or not ctx.previous_return.status == "skipped"


HISTORY = _HistoryGuards
MODE = _ModeGuards
PREVIOUS_TASK = _PreviousTaskGuards



from automyte.automaton import RunContext
from automyte.discovery import File


class flush:
    """Util to force flushing of the file.

    Might be useful if you need to flush the file to the disk, before postprocess.
    """

    def __call__(self, ctx: RunContext, file: File | None):
        if file:
            file.flush()
        else:
            ctx.project.apply_changes()



from pathlib import Path

import pytest

from automyte import AutomatonRunResult, InFileHistory


class TestInFileHistoryRead:
    def test_read_properly_parses_history_file(self, tmp_csv_file):
        data = [
            ["automaton", "project", "status", "error"],
            ["auto1", "proj1", "success", ""],
            ["auto1", "proj2", "skipped", ""],
            ["auto2", "proj1", "fail", "oops"],
        ]

        history = InFileHistory(filename=tmp_csv_file(data))
        result1 = history.read(automaton_name="auto1")
        result2 = history.read(automaton_name="auto2")

        assert result1["proj1"] == AutomatonRunResult(status="success")
        assert result1["proj2"] == AutomatonRunResult(status="skipped")
        assert result2["proj1"] == AutomatonRunResult(status="fail", error="oops")

    def test_read_from_empty_file(self, tmp_csv_file):
        data = [[]]
        result = InFileHistory(filename=tmp_csv_file(data)).read("whatever")

        assert result == {}

    def test_read_from_file_without_necessary_automaton(self, tmp_csv_file):
        data = [
            ["automaton", "project", "status", "error"],
            ["auto9", "proj1", "success", ""],
        ]
        result = InFileHistory(filename=tmp_csv_file(data)).read("auto1")

        assert result == {}

    def test_read_will_raise_if_dir_doesnt_exist(self):
        with pytest.raises(ValueError, match=".*does not exist*"):
            InFileHistory(filename="some/nonexistant/like/definitely/path/").read("whatever")

    def test_read_will_create_empty_default_file_in_existing_dir(self, tmp_local_project):
        dir = tmp_local_project(structure={"src": {}})

        result = InFileHistory(filename=f"{dir}/src/").read("whatever")

        assert result == {}
        assert (Path(dir) / "src" / "automyte_history.csv").exists()

    def test_read_will_raise_if_file_parent_dir_doesnt_exist(self, tmp_local_project):
        dir = tmp_local_project(structure={"src": {}})

        with pytest.raises(ValueError, match=".*does not exist*"):
            InFileHistory(filename=Path(dir) / "src" / "nonexistant_folder" / "history.csv").read("whatever")

    @pytest.mark.parametrize("filepath", [None, "./", "current", "local"])
    def test_read_will_use_local_call_dir_if_no_filepath_is_given(self, filepath):
        exptected_path = Path.cwd() / "automyte_history.csv"
        try:
            InFileHistory(filename=filepath).read("whatever")
            assert exptected_path.exists()

        finally:
            exptected_path.unlink()


class TestInFileHistoryGetStatus:
    def test_correct_retrieval(self, tmp_csv_file):
        data = [
            ["automaton", "project", "status", "error"],
            ["auto", "proj1", "success", ""],
            ["auto", "proj2", "skipped", ""],
            ["auto", "proj3", "fail", "oops"],
        ]
        result1 = InFileHistory(filename=tmp_csv_file(data)).get_status(automaton_name="auto", project_id="proj1")
        result2 = InFileHistory(filename=tmp_csv_file(data)).get_status(automaton_name="auto", project_id="proj2")
        result3 = InFileHistory(filename=tmp_csv_file(data)).get_status(automaton_name="auto", project_id="proj3")

        assert result1 == AutomatonRunResult("success")
        assert result2 == AutomatonRunResult("skipped")
        assert result3 == AutomatonRunResult("fail", error="oops")

    def test_retrieval_for_new_project(self, tmp_csv_file):
        data = [
            ["automaton", "project", "status", "error"],
        ]
        result = InFileHistory(filename=tmp_csv_file(data)).get_status(automaton_name="auto", project_id="proj1")

        assert result == AutomatonRunResult("new")

    def test_retrieval_for_the_first_run(self, tmp_local_project):
        dir = tmp_local_project(structure={"src": {}})

        # In the first run, file might not exist, making sure we just get new run.
        result = InFileHistory(filename=f"{dir}/src").get_status(automaton_name="auto", project_id="proj1")

        assert result == AutomatonRunResult("new")


class TestInFileHistorySetStatus:
    def test_plain_set_in_existing_new_file(self, tmp_csv_file):
        history = InFileHistory(filename=tmp_csv_file([[]]))

        history.set_status(automaton_name="auto", project_id="proj1", status=AutomatonRunResult(status="success"))

        assert history.get_status("auto", "proj1") == AutomatonRunResult("success")

    def test_override_of_previous_status(self, tmp_csv_file):
        data = [
            ["automaton", "project", "status", "error"],
            ["auto", "proj1", "skipped", ""],
        ]
        history = InFileHistory(filename=tmp_csv_file(data))

        history.set_status(
            automaton_name="auto", project_id="proj1", status=AutomatonRunResult(status="fail", error="whoops")
        )

        assert history.get_status("auto", "proj1") == AutomatonRunResult("fail", error="whoops")

    def test_override_doesnt_touch_other_projects(self, tmp_csv_file):
        data = [
            ["automaton", "project", "status", "error"],
            ["auto", "proj2", "skipped", ""],
            ["auto", "proj3", "success", ""],
            ["auto", "proj4", "fail", "oops"],
            ["auto2", "proj1", "fail", "oops"],
        ]
        history = InFileHistory(filename=tmp_csv_file(data))

        history.set_status(automaton_name="auto", project_id="proj1", status=AutomatonRunResult("success"))

        assert history.get_status("auto", "proj2") == AutomatonRunResult("skipped")
        assert history.get_status("auto", "proj3") == AutomatonRunResult("success")
        assert history.get_status("auto", "proj4") == AutomatonRunResult("fail", "oops")
        assert history.get_status("auto2", "proj1") == AutomatonRunResult("fail", "oops")

    def test_overrides_file_contents(self, tmp_csv_file):
        data = [
            ["automaton", "project", "status", "error"],
            ["auto", "proj1", "skipped", ""],
        ]
        filename = tmp_csv_file(data)
        InFileHistory(filename).set_status(
            automaton_name="auto", project_id="proj1", status=AutomatonRunResult("success")
        )

        separate_history_instance = InFileHistory(filename=filename)
        assert separate_history_instance.get_status("auto", "proj1") == AutomatonRunResult("success")



import typing as t
from dataclasses import dataclass

RunStatus: t.TypeAlias = t.Literal["fail", "success", "skipped", "running", "new"]
ProjectID: t.TypeAlias = str


@dataclass
class AutomatonRunResult:
    status: RunStatus
    error: str | None = None



import contextlib

import pytest

from automyte import Automaton
from automyte.automaton.flow import TasksFlow
from automyte.config import Config
from automyte.discovery import ProjectExplorer
from automyte.discovery.file.os_file import OSFile
from automyte.history.in_memory import InMemoryHistory
from automyte.history.types import AutomatonRunResult
from automyte.project.project import Project
from automyte.vcs import VCS


class DummyExplorer(ProjectExplorer):
    def get_rootdir(self) -> str:
        return "smth"

    def set_rootdir(self, newdir: str):
        return "smth"

    def flush(self):
        self.flushed = True

    def explore(self):
        yield OSFile(fullname="smth/hello.txt")


class DummyVCS(VCS):
    def run(self, *args):
        return

    @contextlib.contextmanager
    def preserve_state(self, config):
        self.preserve_state_called = True
        yield "newdir"


class TestAutomatonInit:
    def test_accepts_plain_list_of_tasks(self):
        dummy_task = lambda ctx, file: ...
        flow = Automaton(name="auto", projects=[], tasks=[dummy_task, dummy_task]).flow

        assert isinstance(flow, TasksFlow)
        assert len(flow.preprocess_tasks) == 0
        assert len(flow.postprocess_tasks) == 0

        assert flow.tasks[0] is dummy_task
        assert flow.tasks[1] is dummy_task

    def test_generates_projects_from_uri(self, tmp_local_project):
        dir1 = tmp_local_project({})
        dir2 = tmp_local_project({})
        projects = Automaton(
            "auto",
            projects=[dir1, Project("proj1", rootdir=dir2)],
            tasks=[],
        ).projects

        assert len(projects) == 2
        assert projects[0].rootdir == dir1


class TestAutomatonTargetting:
    @pytest.mark.parametrize(
        "target, expected_projects",
        [
            ("all", ["proj_new", "proj_fail", "proj_success", "proj_skipped", "proj_not_in_history", "custom_id"]),
            ("new", ["proj_new", "proj_not_in_history", "custom_id"]),
            ("failed", ["proj_fail"]),
            ("skipped", ["proj_skipped"]),
            ("successful", ["proj_success"]),
        ],
    )
    def test_appropriate_targetting_filtering(self, target, expected_projects):
        ran_for = []
        mark_project = lambda ctx, file: ran_for.append(ctx.project.project_id)

        history = InMemoryHistory()
        history.set_status("auto", "proj_new", AutomatonRunResult(status="new"))
        history.set_status("auto", "proj_fail", AutomatonRunResult(status="fail"))
        history.set_status("auto", "proj_success", AutomatonRunResult(status="success"))
        history.set_status("auto", "proj_skipped", AutomatonRunResult(status="skipped"))
        history.set_status("auto", "custom_id", AutomatonRunResult(status="new"))

        config = Config.get_default()
        config.target = target

        Automaton(
            "auto",
            history=history,
            config=config,
            projects=[
                Project("proj_new", explorer=DummyExplorer(), vcs=DummyVCS()),
                Project("proj_fail", explorer=DummyExplorer(), vcs=DummyVCS()),
                Project("proj_success", explorer=DummyExplorer(), vcs=DummyVCS()),
                Project("proj_skipped", explorer=DummyExplorer(), vcs=DummyVCS()),
                Project("proj_not_in_history", explorer=DummyExplorer(), vcs=DummyVCS()),
                Project("custom_id", explorer=DummyExplorer(), vcs=DummyVCS()),
            ],
            tasks=TasksFlow(postprocess=[mark_project]),
        ).run()

        assert sorted(ran_for) == sorted(expected_projects)

    def test_targetting_by_target_id(self):
        ran_for = []
        mark_project = lambda ctx, file: ran_for.append(ctx.project.project_id)

        config = Config.get_default()
        config.target = "somecustomid"

        Automaton(
            "auto",
            config=config,
            projects=[Project("somecustomid", explorer=DummyExplorer(), vcs=DummyVCS())],
            tasks=TasksFlow(postprocess=[mark_project]),
        ).run()

        assert ran_for == ["somecustomid"]


class TestAutomatonUpdatesHistory:
    def test_updates_history(self, tmp_local_project):
        successful_task = lambda ctx, file: True
        history = InMemoryHistory()

        Automaton(
            "auto",
            history=history,
            projects=[Project("proj1", explorer=DummyExplorer(), vcs=DummyVCS())],
            tasks=[successful_task],
        ).run()

        assert history.get_status("auto", "proj1").status == "success"

    def test_updates_history_for_fail(self, tmp_local_project):
        history = InMemoryHistory()

        def failed_task(ctx, file):
            raise Exception("oops")

        Automaton(
            "auto",
            history=history,
            projects=[Project("proj1", explorer=DummyExplorer(), vcs=DummyVCS())],
            tasks=[failed_task],
        ).run()

        assert history.get_status("auto", "proj1") == AutomatonRunResult(status="fail", error="oops")



import typing as t
from dataclasses import dataclass

from automyte.discovery import File

if t.TYPE_CHECKING:
    from .run_context import RunContext

InstructionForAutomaton: t.TypeAlias = t.Literal["abort", "skip", "continue"]


@dataclass
class TaskReturn:
    value: t.Any = None
    instruction: "InstructionForAutomaton" = "continue"
    status: t.Literal["processed", "skipped", "errored"] = "processed"


BaseTask: t.TypeAlias = t.Callable[["RunContext", File | None], TaskReturn | t.Any]
FileTask: t.TypeAlias = t.Callable[["RunContext", File], TaskReturn | t.Any]



from .base import Config
from .main import ExitCode, console_main, main  # noqa: TID252
from .vcs import SupportedVCS, VCSConfig

__all__ = [
    "ExitCode",
    "Config",
    "VCSConfig",
    "SupportedVCS",
    "console_main",
    "main",
]



from __future__ import annotations

import contextlib
import logging
import typing as t
import uuid
from pathlib import Path

from automyte.config import VCSConfig
from automyte.discovery import File, Filter
from automyte.utils import bash
from automyte.utils.random import random_hash

from .base import VCS, VCSCmdBuilder, VCSException

logger = logging.getLogger(__name__)


class Git(VCS):
    def __init__(
        self,
        rootdir: str,
        preferred_workflow: t.Literal["rebase", "merge"] = "rebase",
        remote: str = "origin",
    ) -> None:
        self.preferred_workflow = preferred_workflow
        self.remote = remote
        self.original_rootdir = rootdir
        self.workdir = rootdir

    def run(self, *subcommand_with_flags: str):
        return self._exec(subcommand_with_flags[0], *subcommand_with_flags[1:])

    @contextlib.contextmanager
    def preserve_state(self, config: VCSConfig):
        if config.dont_disrupt_prior_state:
            relative_worktree_path = f"./auto_{random_hash()}"
            result = bash.execute(
                VCSCmdBuilder("git")
                .cmd("worktree")
                .in_dir(self.original_rootdir)
                .args("add", "-b", config.work_branch)
                .args(relative_worktree_path)
                .to_cmd()
            )
            if result.status == "fail":
                logger.error(
                    "[Git]: Failed to create worktree at %s for %s branch:\n%s",
                    relative_worktree_path,
                    config.work_branch,
                    result.output,
                )
                raise VCSException("[Git]: Failed to create worktree")

            self.workdir = str(Path(self.original_rootdir) / relative_worktree_path)
            yield str(self.workdir)

            result = bash.execute(
                VCSCmdBuilder("git")
                .cmd("worktree")
                .in_dir(self.original_rootdir)
                .args("remove", "-f", relative_worktree_path)
                .to_cmd()
            )
            if result.status == "fail":
                logger.warning("[Git]: Failed to remove worktree %s:\n%s", relative_worktree_path, result.output)
            self.workdir = self.original_rootdir

        else:
            yield self.original_rootdir

    def _exec(self, cmd: str, *flags: str):
        return bash.execute(VCSCmdBuilder("git").cmd(cmd).in_dir(self.workdir).args(*flags).to_cmd())
